# -*- coding: utf-8 -*-
# pylint: disable=C0103

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

import numpy as np
from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC

from plotpy import io
from plotpy._scaler import INTERP_AA
from plotpy.config import _
from plotpy.constants import LUTAlpha
from plotpy.coords import canvas_to_axes, pixelround
from plotpy.interfaces import (
    IBaseImageItem,
    IBasePlotItem,
    IColormapImageItemType,
    ICSImageItemType,
    IExportROIImageItemType,
    IHistDataSource,
    IImageItemType,
    ISerializableType,
    ITrackableItemType,
    IVoiImageItemType,
)
from plotpy.items.image.base import RawImageItem
from plotpy.items.image.filter import XYImageFilterItem, to_bins
from plotpy.mathutils.arrayfuncs import get_nan_range
from plotpy.styles.image import ImageParam, RGBImageParam, XYImageParam

if TYPE_CHECKING:
    import guidata.io
    import qwt.scale_map
    from qtpy.QtCore import QPointF, QRectF
    from qtpy.QtGui import QColor, QPainter

    from plotpy.interfaces import IItemType
    from plotpy.widgets.colormap.widget import EditableColormap

try:
    from plotpy._scaler import INTERP_NEAREST, _scale_rect, _scale_xy
except ImportError:
    print(
        ("Module 'plotpy.items.image': missing C extension"),
        file=sys.stderr,
    )
    print(
        ("try running :python setup.py build_ext --inplace -c mingw32"),
        file=sys.stderr,
    )
    raise


# ==============================================================================
# Image item
# ==============================================================================
class ImageItem(RawImageItem):
    """Image item

    Args:
        data: 2D NumPy array
        param: Image parameters
    """

    _can_move = True
    _can_resize = True

    __implements__ = (
        IBasePlotItem,
        IBaseImageItem,
        IHistDataSource,
        IVoiImageItemType,
        IExportROIImageItemType,
    )

    def __init__(
        self, data: np.ndarray | None = None, param: ImageParam | None = None
    ) -> None:
        self.xmin = None
        self.xmax = None
        self.ymin = None
        self.ymax = None
        self._log_data = None
        self._lin_lut_range = None
        self._is_zaxis_log = False
        super().__init__(data=data, param=param)

    # ---- BaseImageItem API ---------------------------------------------------
    def get_default_param(self) -> ImageParam:
        """Return instance of the default image param DataSet"""
        return ImageParam(_("Image"))

    # ---- Serialization methods -----------------------------------------------
    def __reduce__(self) -> tuple:
        """Reduce object to be pickled"""
        fname = self.get_filename()
        if fname is None:
            fn_or_data = self.data
        else:
            fn_or_data = fname
        (xmin, xmax), (ymin, ymax) = self.get_xdata(), self.get_ydata()
        state = (
            self.param,
            self.get_lut_range(),
            fn_or_data,
            self.z(),
            xmin,
            xmax,
            ymin,
            ymax,
        )
        res = (self.__class__, (), state)
        return res

    def __setstate__(self, state: tuple) -> None:
        """Set object state from pickled state"""
        param, lut_range, fn_or_data, z, xmin, xmax, ymin, ymax = state
        self.set_xdata(xmin, xmax)
        self.set_ydata(ymin, ymax)
        self.param = param
        if isinstance(fn_or_data, str):
            self.set_filename(fn_or_data)
            self.load_data()
        elif fn_or_data is not None:  # should happen only with previous API
            self.set_data(fn_or_data)
        self.set_lut_range(lut_range)
        self.setZ(z)
        self.param.update_item(self)

    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        super().serialize(writer)
        (xmin, xmax), (ymin, ymax) = self.get_xdata(), self.get_ydata()
        writer.write(xmin, group_name="xmin")
        writer.write(xmax, group_name="xmax")
        writer.write(ymin, group_name="ymin")
        writer.write(ymax, group_name="ymax")

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        super().deserialize(reader)
        for attr in ("xmin", "xmax", "ymin", "ymax"):
            # Note: do not be tempted to write the symetric code in `serialize`
            # because calling `get_xdata` and `get_ydata` is necessary
            setattr(self, attr, reader.read(attr, func=reader.read_float))

    # ---- Public API ----------------------------------------------------------
    def get_xdata(self, aligned=True) -> tuple[float, float]:
        """Get X data range

        Args:
            aligned: True if aligned (Default value = True)

        Returns:
            (xmin, xmax) tuple
        """
        xmin, xmax = self.xmin, self.xmax
        if xmin is None:
            xmin = 0.0
        if xmax is None:
            xmax = self.data.shape[1]
        if aligned:
            dx = 0.5 * (xmax - xmin) / self.data.shape[1]
            xmin -= dx
            xmax -= dx
        return xmin, xmax

    def get_ydata(self, aligned=True) -> tuple[float, float]:
        """Get Y data range

        Args:
            aligned: True if aligned (Default value = True)

        Returns:
            (ymin, ymax) tuple
        """
        ymin, ymax = self.ymin, self.ymax
        if ymin is None:
            ymin = 0.0
        if ymax is None:
            ymax = self.data.shape[0]
        if aligned:
            dy = 0.5 * (ymax - ymin) / self.data.shape[0]
            ymin -= dy
            ymax -= dy
        return ymin, ymax

    def set_xdata(self, xmin: float | None = None, xmax: float | None = None) -> None:
        """Set X data range

        Args:
            xmin: Minimum X value
            xmax: Maximum X value
        """
        self.xmin, self.xmax = xmin, xmax

    def set_ydata(self, ymin: float | None = None, ymax: float | None = None) -> None:
        """Set Y data range

        Args:
            ymin: Minimum Y value
            ymax: Maximum Y value
        """
        self.ymin, self.ymax = ymin, ymax

    def update_bounds(self) -> None:
        """Update image bounds to fit image shape"""
        if self.data is None:
            return
        (xmin, xmax), (ymin, ymax) = self.get_xdata(), self.get_ydata()
        self.bounds = QC.QRectF(QC.QPointF(xmin, ymin), QC.QPointF(xmax, ymax))

    def get_zaxis_log_state(self):
        """Reimplement image.ImageItem method"""
        return self._is_zaxis_log

    def set_zaxis_log_state(self, state):
        """Reimplement image.ImageItem method"""
        self._is_zaxis_log = state
        plot = self.plot()
        if state:
            self._lin_lut_range = self.get_lut_range()
            if self._log_data is None:
                self._log_data = np.array(np.log10(self.data.clip(1)), dtype=np.float64)
            self.set_lut_range(get_nan_range(self._log_data))
            dtype = self._log_data.dtype
        else:
            self._log_data = None
            self.set_lut_range(self._lin_lut_range)
            dtype = self.data.dtype
        if self.interpolate[0] == INTERP_AA:
            self.interpolate = (INTERP_AA, self.interpolate[1].astype(dtype))
        plot.update_colormap_axis(self)

    # ---- BaseImageItem API ---------------------------------------------------
    def get_pixel_coordinates(self, xplot: float, yplot: float) -> tuple[float, float]:
        """Get pixel coordinates from plot coordinates

        Args:
            xplot: X plot coordinate
            yplot: Y plot coordinate

        Returns:
            Pixel coordinates
        """
        (xmin, xmax), (ymin, ymax) = self.get_xdata(), self.get_ydata()
        xpix = self.data.shape[1] * (xplot - xmin) / float(xmax - xmin)
        ypix = self.data.shape[0] * (yplot - ymin) / float(ymax - ymin)
        return xpix, ypix

    def get_plot_coordinates(self, xpixel: float, ypixel: float) -> tuple[float, float]:
        """Get plot coordinates from pixel coordinates

        Args:
            xpixel: X pixel coordinate
            ypixel: Y pixel coordinate

        Returns:
            Plot coordinates
        """
        (xmin, xmax), (ymin, ymax) = self.get_xdata(), self.get_ydata()
        xplot = xmin + (xmax - xmin) * xpixel / float(self.data.shape[1])
        yplot = ymin + (ymax - ymin) * ypixel / float(self.data.shape[0])
        return xplot, yplot

    def get_x_values(self, i0: int, i1: int) -> np.ndarray:
        """Get X values from pixel indexes

        Args:
            i0: First index
            i1: Second index

        Returns:
            X values corresponding to the given pixel indexes
        """
        xmin, xmax = self.get_xdata()

        def xfunc(index):
            return xmin + (xmax - xmin) * index / float(self.data.shape[1])

        return np.linspace(xfunc(i0), xfunc(i1), i1 - i0, endpoint=False)

    def get_y_values(self, j0: int, j1: int) -> np.ndarray:
        """Get Y values from pixel indexes

        Args:
            j0: First index
            j1: Second index

        Returns:
            Y values corresponding to the given pixel indexes
        """
        ymin, ymax = self.get_ydata()

        def yfunc(index):
            return ymin + (ymax - ymin) * index / float(self.data.shape[0])

        return np.linspace(yfunc(j0), yfunc(j1), j1 - j0, endpoint=False)

    def get_closest_coordinates(self, x: float, y: float) -> tuple[float, float]:
        """
        Get the closest coordinates to the given point

        Args:
            x: X coordinate
            y: Y coordinate

        Returns:
            tuple[float, float]: Closest coordinates
        """
        xmin, xmax = self.get_xdata(aligned=False)
        ymin, ymax = self.get_ydata(aligned=False)
        i, j = self.get_closest_indexes(x, y)
        xpix = np.linspace(xmin, xmax, self.data.shape[1] + 1)
        ypix = np.linspace(ymin, ymax, self.data.shape[0] + 1)
        return xpix[i], ypix[j]

    def _rescale_src_rect(
        self, src_rect: tuple[float, float, float, float]
    ) -> tuple[float, float, float, float]:
        """Rescale source rectangle

        Args:
            src_rect: Source rectangle

        Returns:
            Rescaled source rectangle
        """
        sxl, syt, sxr, syb = src_rect
        xl, yt, xr, yb = self.boundingRect().getCoords()
        H, W = self.data.shape[:2]
        if xr - xl != 0:
            x0 = W * (sxl - xl) / (xr - xl)
            x1 = W * (sxr - xl) / (xr - xl)
        else:
            x0 = x1 = 0
        if yb - yt != 0:
            y0 = H * (syt - yt) / (yb - yt)
            y1 = H * (syb - yt) / (yb - yt)
        else:
            y0 = y1 = 0
        return x0, y0, x1, y1

    def draw_image(
        self,
        painter: QPainter,
        canvasRect: QRectF,
        src_rect: tuple[float, float, float, float],
        dst_rect: tuple[float, float, float, float],
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
    ) -> None:
        """Draw image

        Args:
            painter: Painter
            canvasRect: Canvas rectangle
            src_rect: Source rectangle
            dst_rect: Destination rectangle
            xMap: X axis scale map
            yMap: Y axis scale map
        """
        if self.data is None:
            return
        if self.warn_if_non_linear_scale(painter, canvasRect):
            return

        src2 = self._rescale_src_rect(src_rect)
        dst_rect = tuple([int(i) for i in dst_rect])

        # Not the most efficient way to do it, but it works...
        # --------------------------------------------------------------------------
        if self.get_zaxis_log_state():
            data = self._log_data
        else:
            data = self.data
        # --------------------------------------------------------------------------

        try:
            dest = _scale_rect(
                data, src2, self._offscreen, dst_rect, self.lut, self.interpolate
            )
        except ValueError:
            # This exception is raised when zooming unreasonably inside a pixel
            return
        qrect = QC.QRectF(QC.QPointF(dest[0], dest[1]), QC.QPointF(dest[2], dest[3]))
        painter.drawImage(qrect, self._image, qrect)

    def export_roi(
        self,
        src_rect: tuple[float, float, float, float],
        dst_rect: tuple[float, float, float, float],
        dst_image: np.ndarray,
        apply_lut: bool = False,
        apply_interpolation: bool = False,
        original_resolution: bool = False,
        force_interp_mode: str | None = None,
        force_interp_size: int | None = None,
    ) -> None:
        """
        Export a rectangular area of the image to another image

        Args:
            src_rect: Source rectangle
            dst_rect: Destination rectangle
            dst_image: Destination image
            apply_lut: Apply lut (Default value = False)
            apply_interpolation: Apply interpolation (Default value = False)
            original_resolution: Original resolution (Default value = False)
            force_interp_mode: Force interpolation mode (Default value = None)
            force_interp_size: Force interpolation size (Default value = None)
        """
        if apply_lut:
            a, b, _bg, _cmap = self.lut
        else:
            a, b = 1.0, 0.0
        interp = self.interpolate if apply_interpolation else (INTERP_NEAREST,)
        rescaled_src = self._rescale_src_rect(src_rect)
        _scale_rect(
            self.data,
            rescaled_src,
            dst_image,
            dst_rect,
            (a, b, None),
            interp,
        )

    def move_local_point_to(self, handle: int, pos: QPointF, ctrl: bool = None) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        nx, ny = canvas_to_axes(self, pos)
        xmin, xmax = self.get_xdata()
        ymin, ymax = self.get_ydata()

        handle_x = [xmin, xmax, xmax, xmin][handle]
        handle_y = [ymin, ymin, ymax, ymax][handle]
        sign_xmin = [1, -1, -1, 1][handle]
        sign_xmax = [-1, 1, 1, -1][handle]
        sign_ymin = [1, 1, -1, -1][handle]
        sign_ymax = [-1, -1, 1, 1][handle]

        delta_x = nx - handle_x
        delta_y = ny - handle_y

        self.set_xdata(xmin + delta_x * sign_xmin, xmax + delta_x * sign_xmax)
        self.set_ydata(ymin + delta_y * sign_ymin, ymax + delta_y * sign_ymax)
        self.update_bounds()
        self.update_border()

    def move_local_shape(self, old_pos: QPointF, new_pos: QPointF) -> None:
        """Translate the shape such that old_pos becomes new_pos in canvas coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """
        nx, ny = canvas_to_axes(self, new_pos)
        ox, oy = canvas_to_axes(self, old_pos)
        xmin, xmax = self.get_xdata()
        ymin, ymax = self.get_ydata()
        self.set_xdata(xmin + nx - ox, xmax + nx - ox)
        self.set_ydata(ymin + ny - oy, ymax + ny - oy)
        self.update_bounds()
        self.update_border()
        if self.plot():
            self.plot().SIG_ITEM_MOVED.emit(self, ox, oy, nx, ny)

    def move_with_selection(self, delta_x: float, delta_y: float) -> None:
        """Translate the item together with other selected items

        Args:
            delta_x: Translation in plot coordinates along x-axis
            delta_y: Translation in plot coordinates along y-axis
        """
        xmin, xmax = self.get_xdata()
        ymin, ymax = self.get_ydata()
        self.set_xdata(xmin + delta_x, xmax + delta_x)
        self.set_ydata(ymin + delta_x, ymax + delta_y)
        self.update_bounds()
        self.update_border()


assert_interfaces_valid(ImageItem)


class XYImageItem(RawImageItem):
    """XY image item (non-linear axes)

    Args:
        x: 1D NumPy array of pixel center coordinates, must be increasing
        y: 1D NumPy array of pixel center coordinates, must be increasing
        data: 2D NumPy array
        param: image parameters

    Note:
        Internally, `self.x` and `self.y` store **bin edges** (boundaries between
        pixels), not pixel centers. If input arrays have length nj and ni (matching
        data dimensions), they are converted to bin edges of length nj+1 and ni+1
        using `to_bins()`. Methods that need pixel center coordinates must compute
        them from the stored bin edges.
    """

    __implements__ = (
        IBasePlotItem,
        IBaseImageItem,
        IHistDataSource,
        IVoiImageItemType,
        IExportROIImageItemType,
        ISerializableType,
    )

    def __init__(
        self,
        x: np.ndarray | None = None,
        y: np.ndarray | None = None,
        data: np.ndarray | None = None,
        param: XYImageParam | None = None,
    ) -> None:
        # if x and y are not increasing arrays, sort them and data accordingly
        if x is not None and not np.all(np.diff(x) >= 0):
            x_idx = np.argsort(x)
            x = x[x_idx]
            data = data[:, x_idx]
        if y is not None and not np.all(np.diff(y) >= 0):
            y_idx = np.argsort(y)
            y = y[y_idx]
            data = data[y_idx, :]
        super().__init__(data, param)
        self.x = None
        self.y = None
        if x is not None and y is not None:
            self.set_xy(x, y)

    # ---- Public API ----------------------------------------------------------

    # ---- BaseImageItem API ---------------------------------------------------
    def get_default_param(self) -> XYImageParam:
        """Return instance of the default image param DataSet"""
        return XYImageParam(_("Image"))

    # ---- Pickle methods ------------------------------------------------------
    def __reduce__(self) -> tuple:
        """Reduce object to be pickled"""
        fname = self.get_filename()
        if fname is None:
            fn_or_data = self.data
        else:
            fn_or_data = fname
        state = (self.param, self.get_lut_range(), self.x, self.y, fn_or_data, self.z())
        res = (self.__class__, (), state)
        return res

    def __setstate__(self, state: tuple) -> None:
        """Set object state from pickled state"""
        param, lut_range, x, y, fn_or_data, z = state
        self.param = param
        if isinstance(fn_or_data, str):
            self.set_filename(fn_or_data)
            self.load_data(lut_range)
        elif fn_or_data is not None:  # should happen only with previous API
            self.set_data(fn_or_data, lut_range=lut_range)
        self.set_xy(x, y)
        self.setZ(z)
        self.param.update_item(self)

    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        super().serialize(writer)
        writer.write(self.x, group_name="Xdata")
        writer.write(self.y, group_name="Ydata")

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        super().deserialize(reader)
        x = reader.read(group_name="Xdata", func=reader.read_array)
        y = reader.read(group_name="Ydata", func=reader.read_array)
        self.set_xy(x, y)

    # ---- Public API ----------------------------------------------------------
    def set_xy(self, x: np.ndarray | list[float], y: np.ndarray | list[float]) -> None:
        """Set X and Y data

        Args:
            x: 1D NumPy array of pixel center coordinates, must be increasing
            y: 1D NumPy array of pixel center coordinates, must be increasing

        Raises:
            ValueError: If X or Y are not increasing
            IndexError: If X or Y are not of the right length

        Note:
            Input arrays represent pixel **center** coordinates. Internally, they are
            stored as bin **edges** in `self.x` and `self.y`:
            - If input has length nj (or ni), it's converted to edges via `to_bins()`
            - If input already has length nj+1 (or ni+1), it's used directly as edges
            - The stored edges have length nj+1 and ni+1 (one more than data dimensions)
        """
        ni, nj = self.data.shape
        x = np.array(x, float)
        y = np.array(y, float)
        if not np.all(np.diff(x) >= 0):
            raise ValueError("x must be an increasing 1D array")
        if not np.all(np.diff(y) >= 0):
            raise ValueError("y must be an increasing 1D array")
        if x.shape[0] == nj:
            self.x = to_bins(x)
        elif x.shape[0] == nj + 1:
            self.x = x
        else:
            raise IndexError(f"x must be a 1D array of length {nj:d} or {nj + 1:d}")
        if y.shape[0] == ni:
            self.y = to_bins(y)
        elif y.shape[0] == ni + 1:
            self.y = y
        else:
            raise IndexError(f"y must be a 1D array of length {ni:d} or {ni + 1:d}")
        self.bounds = QC.QRectF(
            QC.QPointF(self.x[0], self.y[0]), QC.QPointF(self.x[-1], self.y[-1])
        )
        self.update_border()

    # --- BaseImageItem API ----------------------------------------------------
    def get_filter(self, filterobj, filterparam):
        """Provides a filter object over this image's content"""
        return XYImageFilterItem(self, filterobj, filterparam)

    def draw_image(
        self,
        painter: QPainter,
        canvasRect: QRectF,
        src_rect: tuple[float, float, float, float],
        dst_rect: tuple[float, float, float, float],
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
    ) -> None:
        """Draw image

        Args:
            painter: Painter
            canvasRect: Canvas rectangle
            src_rect: Source rectangle
            dst_rect: Destination rectangle
            xMap: X axis scale map
            yMap: Y axis scale map
        """
        if self.warn_if_non_linear_scale(painter, canvasRect):
            return
        xytr = self.x, self.y, src_rect
        dst_rect = tuple([int(i) for i in dst_rect])
        dest = _scale_xy(
            self.data, xytr, self._offscreen, dst_rect, self.lut, self.interpolate
        )
        qrect = QC.QRectF(QC.QPointF(dest[0], dest[1]), QC.QPointF(dest[2], dest[3]))
        painter.drawImage(qrect, self._image, qrect)

    def export_roi(
        self,
        src_rect: tuple[float, float, float, float],
        dst_rect: tuple[float, float, float, float],
        dst_image: np.ndarray,
        apply_lut: bool = False,
        apply_interpolation: bool = False,
        original_resolution: bool = False,
        force_interp_mode: str | None = None,
        force_interp_size: int | None = None,
    ) -> None:
        """Export a rectangular area of the image to another image

        Args:
            src_rect: Source rectangle in plot coordinates
            dst_rect: Destination rectangle in pixel coordinates
            dst_image: Destination image array
            apply_lut: Apply LUT (Default value = False)
            apply_interpolation: Apply interpolation (Default value = False)
            original_resolution: Original resolution (Default value = False)
            force_interp_mode: Force interpolation mode (Default value = None)
            force_interp_size: Force interpolation size (Default value = None)
        """
        if apply_lut:
            lut = self.lut
        else:
            a, b = 1.0, 0.0
            lut = (a, b, None, np.zeros((1,), np.uint32))
        interp = self.interpolate if apply_interpolation else (INTERP_NEAREST,)
        xytr = self.x, self.y, src_rect
        _scale_xy(self.data, xytr, dst_image, dst_rect, lut, interp)

    def get_pixel_coordinates(self, xplot: float, yplot: float) -> tuple[float, float]:
        """Get pixel coordinates from plot coordinates

        Args:
            xplot: X plot coordinate
            yplot: Y plot coordinate

        Returns:
            Pixel coordinates (integer pixel indices)
        """
        # self.x and self.y are bin edges. searchsorted finds the right edge index.
        # To get the pixel index, we need the left edge index, which is right_edge - 1
        # But clamp to valid range [0, n-1] where n is number of pixels
        i = self.x.searchsorted(xplot)
        j = self.y.searchsorted(yplot)
        # searchsorted returns the insertion point (right edge index)
        # Pixel index is insertion_point - 1, but clamped to [0, data.shape-1]
        i = max(0, min(i - 1, self.data.shape[1] - 1))
        j = max(0, min(j - 1, self.data.shape[0] - 1))
        return i, j

    def get_plot_coordinates(self, xpixel: float, ypixel: float) -> tuple[float, float]:
        """Get plot coordinates from pixel coordinates

        Args:
            xpixel: X pixel coordinate
            ypixel: Y pixel coordinate

        Returns:
            Plot coordinates (pixel center coordinates)
        """
        # self.x and self.y store bin edges, compute centers
        i = int(pixelround(xpixel))
        j = int(pixelround(ypixel))
        # Protect against out-of-bounds access
        i = max(0, min(i, len(self.x) - 2))
        j = max(0, min(j, len(self.y) - 2))
        x_center = (self.x[i] + self.x[i + 1]) / 2.0
        y_center = (self.y[j] + self.y[j + 1]) / 2.0
        return x_center, y_center

    def get_x_values(self, i0: int, i1: int) -> np.ndarray:
        """Get X values from pixel indexes

        Args:
            i0: First index
            i1: Second index

        Returns:
            X values corresponding to the given pixel indexes (pixel centers)
        """
        # self.x stores bin edges (length nj+1), but we need to return pixel centers
        # Compute centers from edges: center[i] = (edge[i] + edge[i+1]) / 2
        # (Fixes issue #49 - Using cross-section tools on `XYImageItem` images alters
        # the X/Y coordinate arrays)
        return (self.x[i0:i1] + self.x[i0 + 1 : i1 + 1]) / 2.0

    def get_y_values(self, j0: int, j1: int) -> np.ndarray:
        """Get Y values from pixel indexes

        Args:
            j0: First index
            j1: Second index

        Returns:
            Y values corresponding to the given pixel indexes (pixel centers)
        """
        # self.y stores bin edges (length ni+1), but we need to return pixel centers
        # Compute centers from edges: center[j] = (edge[j] + edge[j+1]) / 2
        return (self.y[j0:j1] + self.y[j0 + 1 : j1 + 1]) / 2.0

    def get_closest_coordinates(self, x: float, y: float) -> tuple[float, float]:
        """
        Get the closest coordinates to the given point

        Args:
            x: X coordinate
            y: Y coordinate

        Returns:
            tuple[float, float]: Closest pixel center coordinates
        """
        i, j = self.get_closest_indexes(x, y)
        # self.x and self.y store bin edges, compute centers
        # Protect against out-of-bounds access
        i = max(0, min(i, len(self.x) - 2))
        j = max(0, min(j, len(self.y) - 2))
        x_center = (self.x[i] + self.x[i + 1]) / 2.0
        y_center = (self.y[j] + self.y[j + 1]) / 2.0
        return x_center, y_center

    # ---- IBasePlotItem API ---------------------------------------------------
    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (
            IImageItemType,
            IVoiImageItemType,
            IColormapImageItemType,
            ITrackableItemType,
            ISerializableType,
            ICSImageItemType,
            IExportROIImageItemType,
        )

    # ---- IBaseImageItem API --------------------------------------------------
    def can_sethistogram(self) -> bool:
        """
        Returns True if this item can be associated with a levels histogram

        Returns:
            bool: True if item can be associated with a levels histogram,
             False otherwise
        """
        return True


assert_interfaces_valid(XYImageItem)


# ==============================================================================
# RGB Image with alpha channel
# ==============================================================================
class RGBImageItem(ImageItem):
    """RGB image item

    Args:
        data: 3D NumPy array (shape: NxMx[34] -- 3: RGB, 4: RGBA)
        param: image parameters
    """

    __implements__ = (IBasePlotItem, IBaseImageItem, ISerializableType)

    def __init__(
        self, data: np.ndarray | None = None, param: RGBImageParam | None = None
    ) -> None:
        self.orig_data = None
        super().__init__(data, param)
        self.lut = None

    # ---- BaseImageItem API ---------------------------------------------------
    def get_default_param(self) -> RGBImageParam:
        """Return instance of the default image param DataSet"""
        return RGBImageParam(_("Image"))

    # ---- Public API ----------------------------------------------------------
    def recompute_alpha_channel(self) -> None:
        """Recompute alpha channel"""
        data = self.orig_data
        if self.orig_data is None:
            return
        H, W, NC = data.shape
        R = data[..., 0].astype(np.uint32)
        G = data[..., 1].astype(np.uint32)
        B = data[..., 2].astype(np.uint32)
        use_alpha = self.param.alpha_function != LUTAlpha.NONE.value
        alpha = self.param.alpha
        if NC > 3 and use_alpha:
            A = data[..., 3].astype(np.uint32)
        else:
            A = np.zeros((H, W), np.uint32)
            A[:, :] = int(255 * alpha)
        self.data[:, :] = (A << 24) + (R << 16) + (G << 8) + B

    # --- BaseImageItem API ----------------------------------------------------
    # Override lut/bg handling
    def set_lut_range(self, lut_range: tuple[float, float]) -> None:
        """
        Set the current active lut range

        .. warning::

            This method is not implemented for this item type.
        """
        pass

    def set_background_color(self, qcolor: QColor | str) -> None:
        """Set background color

        Args:
            qcolor: Background color

        .. warning::

            This method is not implemented for this item type.
        """
        self.lut = None

    def set_color_map(
        self, name_or_table: str | EditableColormap, invert: bool | None = None
    ) -> None:
        """Set colormap

        Args:
            name_or_table: Colormap name or colormap
            invert: True to invert colormap, False otherwise (Default value = None,
             i.e. do not change the default behavior)
        """
        self.lut = None

    # ---- RawImageItem API ----------------------------------------------------
    def load_data(self) -> None:
        """Load data from item filename attribute"""
        data = io.imread(self.get_filename(), to_grayscale=False)
        self.set_data(data)

    def set_data(self, data: np.ndarray) -> None:
        """Set image data

        Args:
            data: 2D NumPy array
        """
        H, W, NC = data.shape
        self.orig_data = data
        self.data = np.empty((H, W), np.uint32)
        self.recompute_alpha_channel()
        self.update_bounds()
        self.update_border()
        self.lut = None

    # ---- IBasePlotItem API ---------------------------------------------------
    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (IImageItemType, ITrackableItemType, ISerializableType)

    # ---- IBaseImageItem API --------------------------------------------------
    def can_sethistogram(self) -> bool:
        """
        Returns True if this item can be associated with a levels histogram

        Returns:
            bool: True if item can be associated with a levels histogram,
             False otherwise
        """
        return False


assert_interfaces_valid(RGBImageItem)



from __future__ import annotations

import contextlib
from enum import IntEnum, IntFlag, auto
from functools import partial
from typing import Any, overload

from qtpy.QtCore import QPoint, QSize, Qt, Signal
from qtpy.QtGui import QFontMetrics, QValidator
from qtpy.QtWidgets import (
    QAbstractSlider,
    QApplication,
    QBoxLayout,
    QDoubleSpinBox,
    QHBoxLayout,
    QSlider,
    QSpinBox,
    QStyle,
    QStyleOptionSpinBox,
    QVBoxLayout,
    QWidget,
)

from ._misc import signals_blocked
from ._sliders import QDoubleRangeSlider, QDoubleSlider, QRangeSlider


class LabelPosition(IntEnum):
    NoLabel = 0
    LabelsAbove = auto()
    LabelsBelow = auto()
    LabelsRight = LabelsAbove
    LabelsLeft = LabelsBelow


class EdgeLabelMode(IntFlag):
    NoLabel = 0
    LabelIsRange = auto()
    LabelIsValue = auto()


class _SliderProxy:
    _slider: QSlider

    def value(self) -> int:
        return self._slider.value()

    def setValue(self, value: int) -> None:
        self._slider.setValue(value)

    def sliderPosition(self) -> int:
        return self._slider.sliderPosition()

    def setSliderPosition(self, pos: int) -> None:
        self._slider.setSliderPosition(pos)

    def minimum(self) -> int:
        return self._slider.minimum()

    def setMinimum(self, minimum: int) -> None:
        self._slider.setMinimum(minimum)

    def maximum(self) -> int:
        return self._slider.maximum()

    def setMaximum(self, maximum: int) -> None:
        self._slider.setMaximum(maximum)

    def singleStep(self):
        return self._slider.singleStep()

    def setSingleStep(self, step: int) -> None:
        self._slider.setSingleStep(step)

    def pageStep(self) -> int:
        return self._slider.pageStep()

    def setPageStep(self, step: int) -> None:
        self._slider.setPageStep(step)

    def setRange(self, min: int, max: int) -> None:
        self._slider.setRange(min, max)

    def tickInterval(self) -> int:
        return self._slider.tickInterval()

    def setTickInterval(self, interval: int) -> None:
        self._slider.setTickInterval(interval)

    def tickPosition(self) -> QSlider.TickPosition:
        return self._slider.tickPosition()

    def setTickPosition(self, pos: QSlider.TickPosition) -> None:
        self._slider.setTickPosition(pos)

    def __getattr__(self, name: Any) -> Any:
        return getattr(self._slider, name)


def _handle_overloaded_slider_sig(
    args: tuple, kwargs: dict
) -> tuple[QWidget | None, Qt.Orientation]:
    """Maintaining signature of QSlider.__init__."""
    parent = None
    orientation = Qt.Orientation.Horizontal
    errmsg = (
        "TypeError: arguments did not match any overloaded call:\n"
        "  QSlider(parent: QWidget = None)\n"
        "  QSlider(Qt.Orientation, parent: QWidget = None)"
    )
    if len(args) > 2:
        raise TypeError(errmsg)
    elif len(args) == 2:
        if kwargs:
            raise TypeError(errmsg)
        orientation, parent = args
    elif args:
        if isinstance(args[0], QWidget):
            if kwargs:
                raise TypeError(errmsg)
            parent = args[0]
        else:
            orientation = args[0]
    parent = kwargs.get("parent", parent)
    return parent, orientation


class QLabeledSlider(_SliderProxy, QAbstractSlider):
    editingFinished = Signal()

    _slider_class = QSlider
    _slider: QSlider

    @overload
    def __init__(self, parent: QWidget | None = ...) -> None: ...

    @overload
    def __init__(
        self, orientation: Qt.Orientation, parent: QWidget | None = ...
    ) -> None: ...

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        parent, orientation = _handle_overloaded_slider_sig(args, kwargs)

        super().__init__(parent)
        # accept focus events
        fp = self.style().styleHint(QStyle.StyleHint.SH_Button_FocusPolicy)
        self.setFocusPolicy(Qt.FocusPolicy(fp))

        self._slider = self._slider_class(parent=self)
        self._label = SliderLabel(self._slider, connect=self._setValue, parent=self)
        self._edge_label_mode: EdgeLabelMode = EdgeLabelMode.LabelIsValue

        self._rename_signals()
        self._slider.actionTriggered.connect(self.actionTriggered.emit)
        self._slider.rangeChanged.connect(self._on_slider_range_changed)
        self._slider.sliderMoved.connect(self.sliderMoved.emit)
        self._slider.sliderPressed.connect(self.sliderPressed.emit)
        self._slider.sliderReleased.connect(self.sliderReleased.emit)
        self._slider.valueChanged.connect(self._on_slider_value_changed)
        self._label.editingFinished.connect(self.editingFinished)

        self.setOrientation(orientation)

    # ------------------- public API -------------------

    def setOrientation(self, orientation: Qt.Orientation) -> None:
        """Set orientation, value will be 'horizontal' or 'vertical'."""
        self._slider.setOrientation(orientation)
        marg = (0, 0, 0, 0)
        if orientation == Qt.Orientation.Vertical:
            layout = QVBoxLayout()
            layout.addWidget(self._slider, alignment=Qt.AlignmentFlag.AlignHCenter)
            layout.addWidget(self._label, alignment=Qt.AlignmentFlag.AlignHCenter)
            self._label.setAlignment(Qt.AlignmentFlag.AlignCenter)
            layout.setSpacing(1)
        else:
            if self._edge_label_mode == EdgeLabelMode.NoLabel:
                marg = (0, 0, 5, 0)

            layout = QHBoxLayout()  # type: ignore
            layout.addWidget(self._slider)
            layout.addWidget(self._label)
            self._label.setAlignment(Qt.AlignmentFlag.AlignRight)
            layout.setSpacing(6)

        old_layout = self.layout()
        if old_layout is not None:
            QWidget().setLayout(old_layout)

        layout.setContentsMargins(*marg)
        self.setLayout(layout)

    def edgeLabelMode(self) -> EdgeLabelMode:
        """Return current `EdgeLabelMode`."""
        return self._edge_label_mode

    def setEdgeLabelMode(self, opt: EdgeLabelMode) -> None:
        """Set the `EdgeLabelMode`.

        Parameters
        ----------
        opt : EdgeLabelMode
            To show no label, use `EdgeLabelMode.NoLabel`. To show the value
            of the slider, use `EdgeLabelMode.LabelIsValue`. To show
            `value / maximum`, use
            `EdgeLabelMode.LabelIsValue | EdgeLabelMode.LabelIsRange`.
        """
        if opt is EdgeLabelMode.LabelIsRange:
            raise ValueError(
                "mode must be one of 'EdgeLabelMode.NoLabel' or "
                "'EdgeLabelMode.LabelIsValue' or"
                "'EdgeLabelMode.LabelIsValue | EdgeLabelMode.LabelIsRange'."
            )

        self._edge_label_mode = opt
        if not self._edge_label_mode:
            self._label.hide()
            w = 5 if self.orientation() == Qt.Orientation.Horizontal else 0
            self.layout().setContentsMargins(0, 0, w, 0)
        if opt & EdgeLabelMode.LabelIsValue:
            if self.isVisible():
                self._label.show()
            self._label.setMode(opt)
            self._label.setValue(self._slider.value())
            self.layout().setContentsMargins(0, 0, 0, 0)
        self._on_slider_range_changed(self.minimum(), self.maximum())

        QApplication.processEvents()

    # putting this after labelMode methods for the sake of mypy
    EdgeLabelMode = EdgeLabelMode

    # --------------------- private api --------------------

    def _on_slider_range_changed(self, min_: int, max_: int) -> None:
        slash = " / " if self._edge_label_mode & EdgeLabelMode.LabelIsValue else ""
        if self._edge_label_mode & EdgeLabelMode.LabelIsRange:
            self._label.setSuffix(f"{slash}{max_}")
        self.rangeChanged.emit(min_, max_)

    def _on_slider_value_changed(self, v: Any) -> None:
        self._label.setValue(v)
        self.valueChanged.emit(v)

    def _setValue(self, value: float) -> None:
        """Convert the value from float to int before setting the slider value."""
        self._slider.setValue(int(value))

    def _rename_signals(self) -> None:
        # for subclasses
        pass


class QLabeledDoubleSlider(QLabeledSlider):
    _slider_class = QDoubleSlider
    _slider: QDoubleSlider
    _fvalueChanged = Signal(float)
    _fsliderMoved = Signal(float)
    _frangeChanged = Signal(float, float)

    @overload
    def __init__(self, parent: QWidget | None = ...) -> None: ...

    @overload
    def __init__(
        self, orientation: Qt.Orientation, parent: QWidget | None = ...
    ) -> None: ...

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.setDecimals(2)

    def _setValue(self, value: float) -> None:
        """Convert the value from float to int before setting the slider value."""
        self._slider.setValue(value)

    def _rename_signals(self) -> None:
        self.valueChanged = self._fvalueChanged
        self.sliderMoved = self._fsliderMoved
        self.rangeChanged = self._frangeChanged

    def decimals(self) -> int:
        return self._label.decimals()

    def setDecimals(self, prec: int) -> None:
        self._label.setDecimals(prec)


class QLabeledRangeSlider(_SliderProxy, QAbstractSlider):
    _valueChanged = Signal(tuple)
    editingFinished = Signal()

    _slider_class = QRangeSlider
    _slider: QRangeSlider

    @overload
    def __init__(self, parent: QWidget | None = ...) -> None: ...

    @overload
    def __init__(
        self, orientation: Qt.Orientation, parent: QWidget | None = ...
    ) -> None: ...

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        parent, orientation = _handle_overloaded_slider_sig(args, kwargs)
        super().__init__(parent)
        self._rename_signals()

        self.setAttribute(Qt.WidgetAttribute.WA_ShowWithoutActivating)
        self._handle_labels = []
        self._handle_label_position: LabelPosition = LabelPosition.LabelsAbove

        # for fine tuning label position
        self.label_shift_x = 0
        self.label_shift_y = 0

        self._slider = self._slider_class()
        self._slider.valueChanged.connect(self.valueChanged.emit)
        self._slider.rangeChanged.connect(self.rangeChanged.emit)
        self.sliderMoved = self._slider._slidersMoved

        self._min_label = SliderLabel(
            self._slider,
            alignment=Qt.AlignmentFlag.AlignLeft,
            connect=self._min_label_edited,
        )
        self._max_label = SliderLabel(
            self._slider,
            alignment=Qt.AlignmentFlag.AlignRight,
            connect=self._max_label_edited,
        )
        self._min_label.editingFinished.connect(self.editingFinished)
        self._max_label.editingFinished.connect(self.editingFinished)
        self.setEdgeLabelMode(EdgeLabelMode.LabelIsRange)

        self._slider.valueChanged.connect(self._on_value_changed)
        self._slider.rangeChanged.connect(self._on_range_changed)

        self._on_value_changed(self._slider.value())
        self._on_range_changed(self._slider.minimum(), self._slider.maximum())
        self.setOrientation(orientation)

    # --------------------- public API -------------------

    def handleLabelPosition(self) -> LabelPosition:
        """Return where/whether labels are shown adjacent to slider handles."""
        return self._handle_label_position

    def setHandleLabelPosition(self, opt: LabelPosition) -> None:
        """Set where/whether labels are shown adjacent to slider handles."""
        self._handle_label_position = opt
        for lbl in self._handle_labels:
            if not opt:
                lbl.hide()
            else:
                lbl.show()
        self.setOrientation(self.orientation())

    def edgeLabelMode(self) -> EdgeLabelMode:
        """Return current `EdgeLabelMode`."""
        return self._edge_label_mode

    def setEdgeLabelMode(self, opt: EdgeLabelMode) -> None:
        """Set `EdgeLabelMode`, controls what is shown at the min/max labels."""
        self._edge_label_mode = opt
        if not self._edge_label_mode:
            self._min_label.hide()
            self._max_label.hide()
        else:
            if self.isVisible():
                self._min_label.show()
                self._max_label.show()
            self._min_label.setMode(opt)
            self._max_label.setMode(opt)
        if opt == EdgeLabelMode.LabelIsValue:
            v0, *_, v1 = self._slider.value()
            self._min_label.setValue(v0)
            self._max_label.setValue(v1)
        elif opt == EdgeLabelMode.LabelIsRange:
            self._min_label.setValue(self._slider.minimum())
            self._max_label.setValue(self._slider.maximum())
        QApplication.processEvents()
        self._reposition_labels()

    def setRange(self, min: int, max: int) -> None:
        self._on_range_changed(min, max)

    def setOrientation(self, orientation: Qt.Orientation) -> None:
        """Set orientation, value will be 'horizontal' or 'vertical'."""
        self._slider.setOrientation(orientation)
        if orientation == Qt.Orientation.Vertical:
            layout: QBoxLayout = QVBoxLayout()
            layout.setSpacing(1)
            layout.addWidget(self._max_label)
            layout.addWidget(self._slider)
            layout.addWidget(self._min_label)
            # TODO: set margins based on label width
            if self._handle_label_position == LabelPosition.LabelsLeft:
                marg = (30, 0, 0, 0)
            elif self._handle_label_position == LabelPosition.NoLabel:
                marg = (0, 0, 0, 0)
            else:
                marg = (0, 0, 20, 0)
            layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        else:
            layout = QHBoxLayout()
            layout.setSpacing(7)
            if self._handle_label_position == LabelPosition.LabelsBelow:
                marg = (0, 0, 0, 25)
            elif self._handle_label_position == LabelPosition.NoLabel:
                marg = (0, 0, 0, 0)
            else:
                marg = (0, 25, 0, 0)
            layout.addWidget(self._min_label)
            layout.addWidget(self._slider)
            layout.addWidget(self._max_label)

        # remove old layout
        old_layout = self.layout()
        if old_layout is not None:
            QWidget().setLayout(old_layout)

        self.setLayout(layout)
        layout.setContentsMargins(*marg)
        super().setOrientation(orientation)
        QApplication.processEvents()
        self._reposition_labels()

    def resizeEvent(self, a0) -> None:
        super().resizeEvent(a0)
        self._reposition_labels()

    # putting this after methods above for the sake of mypy
    LabelPosition = LabelPosition
    EdgeLabelMode = EdgeLabelMode

    # ------------- private methods ----------------
    def _rename_signals(self) -> None:
        self.valueChanged = self._valueChanged

    def _reposition_labels(self) -> None:
        if (
            not self._handle_labels
            or self._handle_label_position == LabelPosition.NoLabel
        ):
            return

        horizontal = self.orientation() == Qt.Orientation.Horizontal
        labels_above = self._handle_label_position == LabelPosition.LabelsAbove

        last_edge = None
        for i, label in enumerate(self._handle_labels):
            rect = self._slider._handleRect(i)
            dx = -label.width() / 2
            dy = -label.height() / 2
            if labels_above:
                if horizontal:
                    dy *= 3
                else:
                    dx *= -1
            else:
                if horizontal:
                    dy *= -1
                else:
                    dx *= 3
            pos = self._slider.mapToParent(rect.center())
            pos += QPoint(int(dx + self.label_shift_x), int(dy + self.label_shift_y))
            if last_edge is not None:
                # prevent label overlap
                if horizontal:
                    pos.setX(int(max(pos.x(), last_edge.x() + label.width() / 2 + 12)))
                else:
                    pos.setY(int(min(pos.y(), last_edge.y() - label.height() / 2 - 4)))
            label.move(pos)
            last_edge = pos
            label.clearFocus()
            label.show()
        self.update()

    def _min_label_edited(self, val: float) -> None:
        if self._edge_label_mode == EdgeLabelMode.LabelIsRange:
            self.setMinimum(val)
        else:
            v = list(self._slider.value())
            v[0] = val
            self.setValue(v)
        self._reposition_labels()

    def _max_label_edited(self, val: float) -> None:
        if self._edge_label_mode == EdgeLabelMode.LabelIsRange:
            self.setMaximum(val)
        else:
            v = list(self._slider.value())
            v[-1] = val
            self.setValue(v)
        self._reposition_labels()

    def _on_value_changed(self, v: tuple[int, ...]) -> None:
        if self._edge_label_mode == EdgeLabelMode.LabelIsValue:
            self._min_label.setValue(v[0])
            self._max_label.setValue(v[-1])

        if len(v) != len(self._handle_labels):
            for lbl in self._handle_labels:
                lbl.setParent(None)
                lbl.deleteLater()
            self._handle_labels.clear()
            for n, val in enumerate(self._slider.value()):
                _cb = partial(self._slider.setSliderPosition, index=n)
                s = SliderLabel(self._slider, parent=self, connect=_cb)
                s.editingFinished.connect(self.editingFinished)
                s.setValue(val)
                self._handle_labels.append(s)
        else:
            for val, label in zip(v, self._handle_labels):
                label.setValue(val)
        self._reposition_labels()

    def _on_range_changed(self, min: int, max: int) -> None:
        if (min, max) != (self._slider.minimum(), self._slider.maximum()):
            self._slider.setRange(min, max)
        for lbl in self._handle_labels:
            lbl.setRange(min, max)
        if self._edge_label_mode == EdgeLabelMode.LabelIsRange:
            self._min_label.setValue(min)
            self._max_label.setValue(max)
        self._reposition_labels()

    # def setValue(self, value) -> None:
    #     super().setValue(value)
    #     self.sliderChange(QSlider.SliderValueChange)


class QLabeledDoubleRangeSlider(QLabeledRangeSlider):
    _slider_class = QDoubleRangeSlider
    _slider: QDoubleRangeSlider
    _frangeChanged = Signal(float, float)

    @overload
    def __init__(self, parent: QWidget | None = ...) -> None: ...

    @overload
    def __init__(
        self, orientation: Qt.Orientation, parent: QWidget | None = ...
    ) -> None: ...

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.setDecimals(2)

    def _rename_signals(self) -> None:
        super()._rename_signals()
        self.rangeChanged = self._frangeChanged

    def decimals(self) -> int:
        return self._min_label.decimals()

    def setDecimals(self, prec: int) -> None:
        self._min_label.setDecimals(prec)
        self._max_label.setDecimals(prec)
        for lbl in self._handle_labels:
            lbl.setDecimals(prec)


class SliderLabel(QDoubleSpinBox):
    def __init__(
        self,
        slider: QSlider,
        parent=None,
        alignment=Qt.AlignmentFlag.AlignCenter,
        connect=None,
    ) -> None:
        super().__init__(parent=parent)
        self._slider = slider
        self.setFocusPolicy(Qt.FocusPolicy.ClickFocus)
        self.setMode(EdgeLabelMode.LabelIsValue)
        self.setDecimals(0)

        self.setRange(slider.minimum(), slider.maximum())
        slider.rangeChanged.connect(self._update_size)
        self.setAlignment(alignment)
        self.setButtonSymbols(QSpinBox.ButtonSymbols.NoButtons)
        self.setStyleSheet("background:transparent; border: 0;")
        if connect is not None:
            self.editingFinished.connect(lambda: connect(self.value()))
        self.editingFinished.connect(self._silent_clear_focus)
        self._update_size()

    def setDecimals(self, prec: int) -> None:
        super().setDecimals(prec)
        self._update_size()

    def setValue(self, val: Any) -> None:
        super().setValue(val)
        if self._mode == EdgeLabelMode.LabelIsRange:
            self._update_size()

    def setMaximum(self, max: float) -> None:
        super().setMaximum(max)
        if self._mode == EdgeLabelMode.LabelIsValue:
            self._update_size()

    def setMinimum(self, min: float) -> None:
        super().setMinimum(min)
        if self._mode == EdgeLabelMode.LabelIsValue:
            self._update_size()

    def setMode(self, opt: EdgeLabelMode) -> None:
        # when the edge labels are controlling slider range,
        # we want them to have a big range, but not have a huge label
        self._mode = opt
        if opt == EdgeLabelMode.LabelIsRange:
            self.setMinimum(-9999999)
            self.setMaximum(9999999)
            with contextlib.suppress(Exception):
                self._slider.rangeChanged.disconnect(self.setRange)
        else:
            self.setMinimum(self._slider.minimum())
            self.setMaximum(self._slider.maximum())
            self._slider.rangeChanged.connect(self.setRange)
        self._update_size()

    # --------------- private ----------------

    def _silent_clear_focus(self) -> None:
        with signals_blocked(self):
            self.clearFocus()

    def _update_size(self, *_: Any) -> None:
        # fontmetrics to measure the width of text
        fm = QFontMetrics(self.font())
        h = self.sizeHint().height()
        fixed_content = self.prefix() + self.suffix() + " "

        if self._mode & EdgeLabelMode.LabelIsValue:
            # determine width based on min/max/specialValue
            mintext = self.textFromValue(self.minimum())[:18]
            maxtext = self.textFromValue(self.maximum())[:18]
            w = max(0, _fm_width(fm, mintext + fixed_content))
            w = max(w, _fm_width(fm, maxtext + fixed_content))
            if self.specialValueText():
                w = max(w, _fm_width(fm, self.specialValueText()))
            if self._mode & EdgeLabelMode.LabelIsRange:
                w += 8  # it seems as thought suffix() is not enough
        else:
            w = max(0, _fm_width(fm, self.textFromValue(self.value()))) + 3

        w += 3  # cursor blinking space
        # get the final size hint
        opt = QStyleOptionSpinBox()
        self.initStyleOption(opt)
        size = self.style().sizeFromContents(
            QStyle.ContentsType.CT_SpinBox, opt, QSize(w, h), self
        )
        self.setFixedSize(size)

    def validate(
        self, input_: str | None, pos: int
    ) -> tuple[QValidator.State, str, int]:
        # fake like an integer spinbox
        if input_ and "." in input_ and self.decimals() < 1:
            return QValidator.State.Invalid, input_, len(input_)
        return super().validate(input_, pos)


def _fm_width(fm: QFontMetrics, text: str) -> int:
    if hasattr(fm, "horizontalAdvance"):
        return fm.horizontalAdvance(text)
    return fm.width(text)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Renders a cross section chosen by a cross marker"""

# guitest: show

import numpy as np
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import get_path


def create_window():
    win = make.dialog(
        edit=False,
        toolbar=True,
        wintitle="Cross sections test",
        show_xsection=True,
        show_ysection=True,
        type="image",
        size=(640, 600),
    )
    return win


def test_cross_section():
    """Test cross section"""
    with qt_app_context(exec_loop=True):
        filename = get_path("brain.png")
        win = create_window()
        win.show()
        image = make.image(filename=filename, colormap="bone")
        data2 = np.array(image.data.T[200:], copy=True)
        image2 = make.image(data2, title="Modified", alpha_function="linear")
        plot = win.manager.get_plot()
        plot.add_item(image)
        plot.add_item(image2, z=1)


if __name__ == "__main__":
    test_cross_section()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Large number of annotations test"""

# guitest: show

import timeit

import numpy as np
import pytest
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import data as ptd
from plotpy.tests import vistools as ptv

WIN_REFS = []


def plot(items, title):
    """Plot items"""
    global WIN_REFS
    title = f"{title} - #refs: {len(WIN_REFS) + 1}"
    win = ptv.show_items(items, plot_type="image", wintitle=title, title=title)
    WIN_REFS.append(win)
    return win


def create_polygon_vertices(
    x0: float, y0: float, nb_points: int, radius_mean: float, radius_variation: float
) -> list[tuple[float, float]]:
    """Generate points for a polygon around (x0, y0)

    Args:
        x0: x coordinate of the polygon center
        y0: y coordinate of the polygon center
        nb_points: number of points to generate
        radius_mean: mean radius of the polygon
        radius_variation: variation in radius

    Returns:
        List of (x, y) points representing the polygon vertices
    """
    points = []
    # Calculate number of NaNs to append (random numbers between 0 and nb_points-10):
    num_nans = np.random.randint(0, nb_points - 10)
    for j in range(nb_points - num_nans):
        angle = j * (2 * np.pi / (nb_points - num_nans))
        radius = radius_mean + radius_variation * np.random.rand()
        x = x0 + radius * np.cos(angle)
        y = y0 + radius * np.sin(angle)
        points.append((x, y))
    for _ in range(num_nans):
        points.append((np.nan, np.nan))
    return points


def create_random_polygons(
    size: int, nb_polygons: int, nb_points_per_polygon: int
) -> np.ndarray:
    """Create random polygons

    Args:
        size: size of the area in which to create polygons
        nb_polygons: number of polygons to create
        nb_points_per_polygon: number of points per polygon

    Returns:
        Array of shape (nb_polygons, nb_points_per_polygon, 2)
    """
    polygons = []
    for _ in range(nb_polygons):
        x0 = size * np.random.rand()
        y0 = size * np.random.rand()
        points = create_polygon_vertices(
            x0, y0, nb_points_per_polygon, radius_mean=20, radius_variation=30
        )
        # Append the flattened points:
        polygons.append(points)
    return np.array(polygons)


@pytest.mark.skip(reason="Not relevant in automated test suite")
def test_large_number_of_annotations(measure_execution_time: bool = False):
    """Test large number of annotations

    Args:
        measure_execution_time: if True, measure and print the execution time, then
         quit the application immediately.
    """
    title = "Large number of annotations test"
    size = 2000
    nb_polygons = 100
    nb_points_per_polygon = 120
    with qt_app_context(exec_loop=not measure_execution_time):
        data = ptd.gen_image4(size, size)
        x = np.linspace(0, size - 1, size)
        y = np.linspace(0, size - 1, size)
        item = make.xyimage(x, y, data)
        item.set_readonly(True)
        item.set_selectable(False)
        items = [item]
        polygons = create_random_polygons(size, nb_polygons, nb_points_per_polygon)
        for i, points in enumerate(polygons):
            polygon = make.annotated_polygon(np.array(points), title=f"Polygon {i + 1}")
            polygon.set_closed(False)
            items.append(polygon)
        if measure_execution_time:
            exec_time = timeit.timeit(lambda: plot(items, title), number=10) / 10
            print(
                f"Execution time with {nb_polygons} polygons "
                f"({nb_points_per_polygon} points per polygon): "
                f"{exec_time:.3f} seconds/plot"
            )
        else:
            win = plot(items, title)
            win.register_annotation_tools()


if __name__ == "__main__":
    test_large_number_of_annotations(measure_execution_time=True)



# -*- coding: utf-8 -*-

"""
Cross-sections
^^^^^^^^^^^^^^

The cross-section panels are used to display cross-sections of 2D data, either
along the X or Y axis, or along an oblique rectangle (average of the data along
an oblique rectangle).

.. autoclass:: XCrossSection
.. autoclass:: YCrossSection
.. autoclass:: ObliqueCrossSection
.. autoclass:: LineCrossSection
"""

# pylint: disable=unused-import
# flake8: noqa

from .cswidget import (
    ObliqueCrossSection,
    XCrossSection,
    YCrossSection,
    LineCrossSection,
)



# -*- coding: utf-8 -*-

# pylint: disable=unused-import
# flake8: noqa

from .base import PanelWidget
from .contrastadjustment import ContrastAdjustment
from .csection import (
    ObliqueCrossSection,
    XCrossSection,
    YCrossSection,
    LineCrossSection,
)
from .itemlist import PlotItemList



from __future__ import annotations

from guidata.qthelpers import exec_dialog, qt_app_context

from plotpy.interfaces.items import IImageItemType
from plotpy.tests.unit.utils import create_window
from plotpy.tools import AspectRatioTool


def test_aspect_ratio_tool():
    """Test the aspect ratio tool."""
    with qt_app_context(exec_loop=False):
        win, tool = create_window(AspectRatioTool, active_item_type=IImageItemType)
        plot = win.manager.get_plot()

        initial_aspect_ratio: float = plot.get_aspect_ratio()

        new_ratio = 0.5

        plot.set_aspect_ratio(new_ratio)
        tool.edit_aspect_ratio()
        assert plot.get_aspect_ratio() == new_ratio

        plot.set_aspect_ratio(initial_aspect_ratio)
        tool.edit_aspect_ratio()
        assert plot.get_aspect_ratio() == initial_aspect_ratio

        tool.lock_aspect_ratio(True)
        assert plot.lock_aspect_ratio is True
        tool.lock_aspect_ratio(False)
        assert plot.lock_aspect_ratio is False

        exec_dialog(win)


if __name__ == "__main__":
    test_aspect_ratio_tool()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# pylint: disable=C0103

"""
Image Item builder
------------------

This module provides a set of factory functions to simplify the creation
of image items.
"""

# Note: when adding method to builder classes, please do not forget to update the
# documentation (see builder.rst file). Because of class inheritance, the methods
# are not automatically documented (otherwise, they would be sorted alphabetically,
# due to a limitation of sphinx auto-doc).

from __future__ import annotations

import os.path as osp
from collections.abc import Callable
from typing import TYPE_CHECKING

import numpy  # only to help intersphinx finding numpy doc
import numpy as np

from plotpy import io
from plotpy.config import _, make_title
from plotpy.constants import LUTAlpha
from plotpy.items import (
    ContourItem,
    Histogram2DItem,
    ImageItem,
    MaskedImageItem,
    MaskedXYImageItem,
    QuadGridItem,
    RGBImageItem,
    TrImageItem,
    XYImageItem,
    create_contour_items,
)
from plotpy.styles import (
    Histogram2DParam,
    ImageFilterParam,
    ImageParam,
    MaskedImageParam,
    MaskedXYImageParam,
    QuadGridParam,
    RGBImageParam,
    TrImageParam,
    XYImageParam,
)

if TYPE_CHECKING:
    from plotpy.items.image.filter import ImageFilterItem

IMAGE_COUNT = 0
HISTOGRAM2D_COUNT = 0


class ImageBuilder:
    """Class regrouping a set of factory functions to simplify the creation
    of image items."""

    def __set_image_param(
        self,
        param: ImageParam,
        title: str | None,
        alpha_function: LUTAlpha | str | None,
        alpha: float | None,
        interpolation: str,
        **kwargs,
    ) -> None:
        if title:
            param.label = title
        else:
            global IMAGE_COUNT
            IMAGE_COUNT += 1
            param.label = make_title(_("Image"), IMAGE_COUNT)
        if alpha_function is not None:
            if isinstance(alpha_function, str):
                alpha_function = LUTAlpha.get_member_from_name(alpha_function)
            assert isinstance(alpha_function, LUTAlpha)
            param.alpha_function = alpha_function.value
        if alpha is not None:
            assert 0.0 <= alpha <= 1.0
            param.alpha = alpha
        interp_methods = {"nearest": 0, "linear": 1, "antialiasing": 5}
        param.interpolation = interp_methods[interpolation]
        for key, val in list(kwargs.items()):
            if val is not None:
                setattr(param, key, val)

    def _get_image_data(
        self,
        data: numpy.ndarray,
        filename: str | None,
        title: str | None,
        to_grayscale: bool,
    ) -> tuple[numpy.ndarray, str | None, str | None]:
        if data is None:
            assert filename is not None
            data = io.imread(filename, to_grayscale=to_grayscale)
        if title is None and filename is not None:
            title = osp.basename(filename)
        return data, filename, title

    @staticmethod
    def compute_bounds(
        data: numpy.ndarray,
        pixel_size: float | tuple[float, float],
        center_on: tuple[float, float] | None = None,
    ) -> tuple[float, float, float, float]:
        """Return image bounds from *pixel_size* (scalar or tuple)

        Args:
            data: image data
            pixel_size: pixel size
            center_on: center coordinates. Default is None

        Returns:
            tuple: xmin, xmax, ymin, ymax
        """
        if not isinstance(pixel_size, (tuple, list)):
            pixel_size = [pixel_size, pixel_size]
        dx, dy = pixel_size
        xmin, ymin = 0.0, 0.0
        xmax, ymax = data.shape[1] * dx, data.shape[0] * dy
        if center_on is not None:
            xc, yc = center_on
            dx, dy = 0.5 * (xmax - xmin) - xc, 0.5 * (ymax - ymin) - yc
            xmin -= dx
            xmax -= dx
            ymin -= dy
            ymax -= dy
        return xmin, xmax, ymin, ymax

    def image(
        self,
        data: numpy.ndarray | None = None,
        filename: str | None = None,
        title: str | None = None,
        alpha_function: LUTAlpha | str | None = None,
        alpha: float | None = None,
        background_color: str | None = None,
        colormap: str | None = None,
        xdata: list[float] = [None, None],
        ydata: list[float] = [None, None],
        pixel_size: float | tuple[float, float] | None = None,
        center_on: tuple[float, float] | None = None,
        interpolation: str = "linear",
        eliminate_outliers: float | None = None,
        xformat: str = "%.1f",
        yformat: str = "%.1f",
        zformat: str = "%.1f",
        x: numpy.ndarray | None = None,
        y: numpy.ndarray | None = None,
        lut_range: tuple[float, float] | None = None,
        lock_position: bool = True,
    ) -> ImageItem | XYImageItem | RGBImageItem:
        """Make an image `plot item` from data

        Args:
            data: data. Default is None
            filename: image filename. Default is None
            title: image title. Default is None
            alpha_function: function for LUT alpha channel.
             Default is :py:attr:`.LUTAlpha.NONE` (valid string values are 'none',
             'constant', 'linear', 'sigmoid', 'tanh' and 'step')
            alpha: alpha value. Default is None
            background_color: background color name. Default is None
            colormap: colormap name. Default is None
            xdata: x data. Default is [None, None]
            ydata: y data. Default is [None, None]
            pixel_size: pixel size. Default is None
            center_on: center on. Default is None
            interpolation: interpolation method. Default is 'linear'
            eliminate_outliers: eliminate outliers. Default is None
            xformat: x format. Default is '%.1f'
            yformat: y format. Default is '%.1f'
            zformat: z format. Default is '%.1f'
            x: x data. Default is None
            y: y data. Default is None
            lut_range: LUT range. Default is None
            lock_position: lock position. Default is True

        Returns:
            :py:class:`.ImageItem` object or
            :py:class:`.XYImageItem` object if `x` and `y` are specified or
            :py:class:`.RGBImageItem` object if data has 3 dimensions
        """
        if x is not None or y is not None:
            assert pixel_size is None and center_on is None, (
                "Ambiguous parameters:"
                "both `x`/`y` and `pixel_size`/`center_on`/`xdata`/`ydata`"
                " were specified"
            )
            return self.xyimage(
                x,
                y,
                data,
                title=title,
                alpha_function=alpha_function,
                alpha=alpha,
                background_color=background_color,
                colormap=colormap,
                interpolation=interpolation,
                eliminate_outliers=eliminate_outliers,
                xformat=xformat,
                yformat=yformat,
                zformat=zformat,
                lut_range=lut_range,
                lock_position=lock_position,
            )

        assert isinstance(xdata, (tuple, list)) and len(xdata) == 2
        assert isinstance(ydata, (tuple, list)) and len(ydata) == 2
        param = ImageParam(title=_("Image"), icon="image.png")
        data, filename, title = self._get_image_data(
            data, filename, title, to_grayscale=True
        )

        if isinstance(filename, str) and filename.lower().endswith(".dcm"):
            # pylint: disable=import-outside-toplevel
            # pylint: disable=import-error
            from pydicom import dcmread  # type:ignore

            template = dcmread(filename, stop_before_pixels=True, force=True)
            ipp = getattr(template, "ImagePositionPatient", ["0", "0", "0"])
            pxs = getattr(template, "PixelSpacing", ["1", "1"])
            ipx, ipy = float(ipp[0]), float(ipp[1])
            pixel_size = dy, dx = float(pxs[0]), float(pxs[1])
            xc = (0.5 * data.shape[1] - 1) * dx + ipx
            yc = (0.5 * data.shape[0] - 1) * dy + ipy
            center_on = xc, yc

        if data.ndim == 3:
            return self.rgbimage(
                data=data,
                filename=filename,
                title=title,
                alpha_function=alpha_function,
                alpha=alpha,
            )
        assert data.ndim == 2, "Data must have 2 dimensions"
        if pixel_size is None:
            assert center_on is None, (
                "Argument `pixel_size` must be specified when `center_on`"
            )
            xmin, xmax = xdata
            ymin, ymax = ydata
        else:
            xmin, xmax, ymin, ymax = self.compute_bounds(data, pixel_size, center_on)
        self.__set_image_param(
            param,
            title,
            alpha_function,
            alpha,
            interpolation,
            background=background_color,
            colormap=colormap,
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
            xformat=xformat,
            yformat=yformat,
            zformat=zformat,
            lock_position=lock_position,
        )
        image = ImageItem(data, param)
        image.set_filename(filename)
        if lut_range is not None:
            assert eliminate_outliers is None, (
                "Ambiguous parameters: both `lut_range`"
                " and `eliminate_outliers` were specified"
            )
            image.set_lut_range(lut_range)
        elif eliminate_outliers is not None:
            image.set_lut_threshold(eliminate_outliers)
        return image

    def maskedimage(
        self,
        data: numpy.ndarray | None = None,
        mask: numpy.ndarray | None = None,
        filename: str | None = None,
        title: str | None = None,
        alpha_function: LUTAlpha | str | None = None,
        alpha: float = 1.0,
        xdata: list[float] = [None, None],
        ydata: list[float] = [None, None],
        pixel_size: float | tuple[float, float] | None = None,
        center_on: tuple[float, float] | None = None,
        background_color: str | None = None,
        colormap: str | None = None,
        show_mask: bool = False,
        fill_value: float | None = None,
        interpolation: str = "linear",
        eliminate_outliers: float | None = None,
        xformat: str = "%.1f",
        yformat: str = "%.1f",
        zformat: str = "%.1f",
        x: numpy.ndarray | None = None,
        y: numpy.ndarray | None = None,
        lut_range: tuple[float, float] | None = None,
        lock_position: bool = True,
    ) -> MaskedImageItem | MaskedXYImageItem:
        """Make a masked image `plot item` from data

        Args:
            data: data. Default is None
            mask: mask. Default is None
            filename: image filename. Default is None
            title: image title. Default is None
            alpha_function: function for LUT alpha channel.
             Default is :py:attr:`.LUTAlpha.NONE` (valid string values are 'none',
             'constant', 'linear', 'sigmoid', 'tanh' and 'step')
            alpha: alpha value. Default is 1.0
            xdata: x data. Default is [None, None]
            ydata: y data. Default is [None, None]
            pixel_size: pixel size. Default is None
            center_on: center on. Default is None
            background_color: background color. Default is None
            colormap: colormap. Default is None
            show_mask: show mask. Default is False
            fill_value: fill value. Default is None
            interpolation: interpolation method. Default is 'linear'
            eliminate_outliers: eliminate outliers. Default is None
            xformat: x format. Default is '%.1f'
            yformat: y format. Default is '%.1f'
            zformat: z format. Default is '%.1f'
            x: x data. Default is None
            y: y data. Default is None
            lut_range: LUT range. Default is None
            lock_position: lock position. Default is True

        Returns:
            :py:class:`.MaskedImageItem` object or :py:class:`.MaskedXYImageItem` object
        """
        if x is not None or y is not None:
            assert pixel_size is None and center_on is None, (
                "Ambiguous parameters:"
                "both `x`/`y` and `pixel_size`/`center_on`/`xdata`/`ydata`"
                " were specified"
            )
            return self.maskedxyimage(
                x,
                y,
                data,
                mask=mask,
                title=title,
                alpha_function=alpha_function,
                alpha=alpha,
                background_color=background_color,
                colormap=colormap,
                show_mask=show_mask,
                fill_value=fill_value,
                interpolation=interpolation,
                eliminate_outliers=eliminate_outliers,
                xformat=xformat,
                yformat=yformat,
                zformat=zformat,
                lut_range=lut_range,
                lock_position=lock_position,
            )

        assert isinstance(xdata, (tuple, list)) and len(xdata) == 2
        assert isinstance(ydata, (tuple, list)) and len(ydata) == 2
        param = MaskedImageParam(title=_("Image"), icon="image.png")
        data, filename, title = self._get_image_data(
            data, filename, title, to_grayscale=True
        )
        assert data.ndim == 2, "Data must have 2 dimensions"
        if pixel_size is None:
            assert center_on is None, (
                "Ambiguous parameters: both `center_on`"
                " and `xdata`/`ydata` were specified"
            )
            xmin, xmax = xdata
            ymin, ymax = ydata
        else:
            xmin, xmax, ymin, ymax = self.compute_bounds(data, pixel_size, center_on)
        self.__set_image_param(
            param,
            title,
            alpha_function,
            alpha,
            interpolation,
            background=background_color,
            colormap=colormap,
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
            show_mask=show_mask,
            fill_value=fill_value,
            xformat=xformat,
            yformat=yformat,
            zformat=zformat,
            lock_position=lock_position,
        )
        image = MaskedImageItem(data, mask, param)
        image.set_filename(filename)
        if lut_range is not None:
            assert eliminate_outliers is None, (
                "Ambiguous parameters: both `lut_range`"
                " and `eliminate_outliers` were specified"
            )
            image.set_lut_range(lut_range)
        elif eliminate_outliers is not None:
            image.set_lut_threshold(eliminate_outliers)
        return image

    def maskedxyimage(
        self,
        x: numpy.ndarray,
        y: numpy.ndarray,
        data: numpy.ndarray,
        mask: numpy.ndarray | None = None,
        title: str | None = None,
        alpha_function: LUTAlpha | str | None = None,
        alpha: float = 1.0,
        background_color: str | None = None,
        colormap: str | None = None,
        show_mask: bool = False,
        fill_value: float | None = None,
        interpolation: str = "linear",
        eliminate_outliers: float | None = None,
        xformat: str = "%.1f",
        yformat: str = "%.1f",
        zformat: str = "%.1f",
        lut_range: tuple[float, float] | None = None,
        lock_position: bool = True,
    ) -> MaskedXYImageItem:
        """Make a masked XY image `plot item` from data

        Args:
            x: x data
            y: y data
            data: data
            mask: mask. Default is None
            title: image title. Default is None
            alpha_function: function for LUT alpha channel.
             Default is :py:attr:`.LUTAlpha.NONE` (valid string values are 'none',
             'constant', 'linear', 'sigmoid', 'tanh' and 'step')
            alpha: alpha value. Default is 1.0
            background_color: background color. Default is None
            colormap: colormap. Default is None
            show_mask: show mask. Default is False
            fill_value: fill value. Default is None
            interpolation: interpolation method. Default is 'linear'
            eliminate_outliers: eliminate outliers. Default is None
            xformat: x format. Default is '%.1f'
            yformat: y format. Default is '%.1f'
            zformat: z format. Default is '%.1f'
            lut_range: LUT range. Default is None
            lock_position: lock position. Default is True

        Returns:
            :py:class:`.MaskedXYImageItem` object
        """
        if isinstance(x, (list, tuple)):
            x = np.array(x)
        if isinstance(y, (list, tuple)):
            y = np.array(y)

        param = MaskedXYImageParam(title=_("Image"), icon="image.png")
        assert data.ndim == 2, "Data must have 2 dimensions"

        self.__set_image_param(
            param,
            title,
            alpha_function,
            alpha,
            interpolation,
            background=background_color,
            colormap=colormap,
            show_mask=show_mask,
            fill_value=fill_value,
            xformat=xformat,
            yformat=yformat,
            zformat=zformat,
            lock_position=lock_position,
        )
        image = MaskedXYImageItem(x, y, data, mask, param)
        if lut_range is not None:
            assert eliminate_outliers is None, (
                "Ambiguous parameters: both `lut_range`"
                " and `eliminate_outliers` were specified"
            )
            image.set_lut_range(lut_range)
        elif eliminate_outliers is not None:
            image.set_lut_threshold(eliminate_outliers)
        return image

    def rgbimage(
        self,
        data: numpy.ndarray | None = None,
        filename: str | None = None,
        title: str | None = None,
        alpha_function: LUTAlpha | str | None = None,
        alpha: float = 1.0,
        xdata: list | tuple = [None, None],
        ydata: list | tuple = [None, None],
        pixel_size: float | None = None,
        center_on: tuple | None = None,
        interpolation: str = "linear",
        lock_position: bool = True,
    ) -> RGBImageItem:
        """Make a RGB image `plot item` from data

        Args:
            data: data
            filename: filename. Default is None
            title: image title. Default is None
            alpha_function: function for LUT alpha channel.
             Default is :py:attr:`.LUTAlpha.NONE` (valid string values are 'none',
             'constant', 'linear', 'sigmoid', 'tanh' and 'step')
            alpha: alpha value. Default is 1.0
            xdata: x data. Default is [None, None]
            ydata: y data. Default is [None, None]
            pixel_size: pixel size. Default is None
            center_on: center on. Default is None
            interpolation: interpolation method. Default is 'linear'
            lock_position: lock position. Default is True

        Returns:
            :py:class:`.RGBImageItem` object
        """
        assert isinstance(xdata, (tuple, list)) and len(xdata) == 2
        assert isinstance(ydata, (tuple, list)) and len(ydata) == 2
        param = RGBImageParam(title=_("Image"), icon="image.png")
        data, filename, title = self._get_image_data(
            data, filename, title, to_grayscale=False
        )
        assert data.ndim == 3, "RGB data must have 3 dimensions"
        if pixel_size is None:
            assert center_on is None, (
                "Ambiguous parameters: both `center_on`"
                " and `xdata`/`ydata` were specified"
            )
            xmin, xmax = xdata
            ymin, ymax = ydata
        else:
            xmin, xmax, ymin, ymax = self.compute_bounds(data, pixel_size, center_on)
        self.__set_image_param(
            param,
            title,
            alpha_function,
            alpha,
            interpolation,
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
            lock_position=lock_position,
        )
        image = RGBImageItem(data, param)
        image.set_filename(filename)
        return image

    def quadgrid(
        self,
        X: numpy.ndarray,
        Y: numpy.ndarray,
        Z: numpy.ndarray,
        title: str | None = None,
        alpha_function: LUTAlpha | str | None = None,
        alpha: float | None = None,
        colormap: str | None = None,
        interpolation: str = "linear",
        lock_position: bool = True,
    ) -> QuadGridItem:
        """Make a pseudocolor `plot item` of a 2D array

        Args:
            X: x data
            Y: y data
            Z: data
            title: image title. Default is None
            alpha_function: function for LUT alpha channel.
             Default is :py:attr:`.LUTAlpha.NONE` (valid string values are 'none',
             'constant', 'linear', 'sigmoid', 'tanh' and 'step')
            alpha: alpha value. Default is None
            colormap: colormap. Default is None
            interpolation: interpolation method. Default is 'linear'
            lock_position: lock position. Default is True

        Returns:
            :py:class:`.QuadGridItem` object
        """
        param = QuadGridParam(title=_("Image"), icon="image.png")
        self.__set_image_param(
            param,
            title,
            alpha_function,
            alpha,
            interpolation,
            colormap=colormap,
            lock_position=lock_position,
        )
        image = QuadGridItem(X, Y, Z, param)
        return image

    def pcolor(self, *args, **kwargs) -> QuadGridItem:
        """Make a pseudocolor `plot item` of a 2D array
        based on MATLAB-like syntax

        Args:
            args: non-keyword arguments
            kwargs: keyword arguments

        Returns:
            :py:class:`.QuadGridItem` object

        Examples::
            pcolor(C)
            pcolor(X, Y, C)
        """  # noqa: E501
        if len(args) == 1:
            (Z,) = args
            M, N = Z.shape
            X, Y = np.meshgrid(np.arange(N, dtype=Z.dtype), np.arange(M, dtype=Z.dtype))
        elif len(args) == 3:
            X, Y, Z = args
        else:
            raise RuntimeError("1 or 3 non-keyword arguments expected")
        return self.quadgrid(X, Y, Z, **kwargs)

    def trimage(
        self,
        data: numpy.ndarray | None = None,
        filename: str | None = None,
        title: str | None = None,
        alpha_function: LUTAlpha | str | None = None,
        alpha: float | None = None,
        background_color: str | None = None,
        colormap: str | None = None,
        x0: float = 0.0,
        y0: float = 0.0,
        angle: float = 0.0,
        dx: float = 1.0,
        dy: float = 1.0,
        interpolation: str = "linear",
        eliminate_outliers: float | None = None,
        xformat: str = "%.1f",
        yformat: str = "%.1f",
        zformat: str = "%.1f",
        lut_range: tuple[float, float] | None = None,
        lock_position: bool = False,
    ) -> TrImageItem:
        """Make a transformable image `plot item` (image with an arbitrary
        affine transform)

        Args:
            data: data
            filename: filename. Default is None
            title: image title. Default is None
            alpha_function: function for LUT alpha channel.
             Default is :py:attr:`.LUTAlpha.NONE` (valid string values are 'none',
             'constant', 'linear', 'sigmoid', 'tanh' and 'step')
            alpha: alpha value. Default is None
            background_color: background color. Default is None
            colormap: colormap. Default is None
            x0: x position. Default is 0.0
            y0: y position. Default is 0.0
            angle: angle (radians). Default is 0.0
            dx: pixel size along X axis. Default is 1.0
            dy: pixel size along Y axis. Default is 1.0
            interpolation: interpolation method. Default is 'linear'
            eliminate_outliers: eliminate outliers. Default is None
            xformat: x format. Default is '%.1f'
            yformat: y format. Default is '%.1f'
            zformat: z format. Default is '%.1f'
            lut_range: LUT range. Default is None
            lock_position: lock position. Default is False

        Returns:
            :py:class:`.TrImageItem` object
        """
        param = TrImageParam(title=_("Image"), icon="image.png")
        data, filename, title = self._get_image_data(
            data, filename, title, to_grayscale=True
        )
        self.__set_image_param(
            param,
            title,
            alpha_function,
            alpha,
            interpolation,
            background=background_color,
            colormap=colormap,
            pos_x0=x0,
            pos_y0=y0,
            angle=angle,
            dx=dx,
            dy=dy,
            xformat=xformat,
            yformat=yformat,
            zformat=zformat,
            lock_position=lock_position,
        )
        image = TrImageItem(data, param)
        image.set_filename(filename)
        if lut_range is not None:
            assert eliminate_outliers is None, (
                "Ambiguous parameters: both `lut_range`"
                " and `eliminate_outliers` were specified"
            )
            image.set_lut_range(lut_range)
        elif eliminate_outliers is not None:
            image.set_lut_threshold(eliminate_outliers)
        return image

    def xyimage(
        self,
        x: numpy.ndarray,
        y: numpy.ndarray,
        data: numpy.ndarray,
        title: str | None = None,
        alpha_function: LUTAlpha | str | None = None,
        alpha: float | None = None,
        background_color: str | None = None,
        colormap: str | None = None,
        interpolation: str = "linear",
        eliminate_outliers: float | None = None,
        xformat: str = "%.1f",
        yformat: str = "%.1f",
        zformat: str = "%.1f",
        lut_range: tuple[float, float] | None = None,
        lock_position: bool = False,
    ) -> XYImageItem:
        """Make an xyimage `plot item` (image with non-linear X/Y axes) from data

        Args:
            x: X coordinates
            y: Y coordinates
            data: data
            title: image title. Default is None
            alpha_function: function for LUT alpha channel.
             Default is :py:attr:`.LUTAlpha.NONE` (valid string values are 'none',
             'constant', 'linear', 'sigmoid', 'tanh' and 'step')
            alpha: alpha value. Default is None
            background_color: background color. Default is None
            colormap: colormap. Default is None
            interpolation: interpolation method. Default is 'linear'
            eliminate_outliers: eliminate outliers. Default is None
            xformat: x format. Default is '%.1f'
            yformat: y format. Default is '%.1f'
            zformat: z format. Default is '%.1f'
            lut_range: LUT range. Default is None
            lock_position: lock position. Default is True

        Returns:
            :py:class:`.XYImageItem` object
        """
        param = XYImageParam(title=_("Image"), icon="image.png")
        self.__set_image_param(
            param,
            title,
            alpha_function,
            alpha,
            interpolation,
            background=background_color,
            colormap=colormap,
            xformat=xformat,
            yformat=yformat,
            zformat=zformat,
            lock_position=lock_position,
        )
        if isinstance(x, (list, tuple)):
            x = np.array(x)
        if isinstance(y, (list, tuple)):
            y = np.array(y)
        image = XYImageItem(x, y, data, param)
        if lut_range is not None:
            assert eliminate_outliers is None, (
                "Ambiguous parameters: both `lut_range`"
                " and `eliminate_outliers` were specified"
            )
            image.set_lut_range(lut_range)
        elif eliminate_outliers is not None:
            image.set_lut_threshold(eliminate_outliers)
        return image

    def imagefilter(
        self,
        xmin: float,
        xmax: float,
        ymin: float,
        ymax: float,
        imageitem: ImageItem,
        filter: Callable,
        title: str | None = None,
    ) -> ImageFilterItem:
        """Make a rectangular area image filter `plot item`

        Args:
            xmin: xmin
            xmax: xmax
            ymin: ymin
            ymax: ymax
            imageitem: image item
            filter: filter function
            title: filter title. Default is None

        Returns:
            :py:class:`.ImageFilterItem` object
        """
        param = ImageFilterParam(_("Filter"), icon="funct.png")
        param.xmin, param.xmax, param.ymin, param.ymax = xmin, xmax, ymin, ymax
        if title is not None:
            param.label = title
        filt = imageitem.get_filter(filter, param)
        _m, _M = imageitem.get_lut_range()
        filt.set_lut_range([_m, _M])
        return filt

    def contours(
        self,
        Z: np.ndarray,
        levels: float | np.ndarray,
        X: np.ndarray | None = None,
        Y: np.ndarray | None = None,
    ) -> list[ContourItem]:
        """Make a contour curves

        Args:
            Z: The height values over which the contour is drawn.
            levels : Determines the number and positions of the contour lines/regions.
             If a float, draw contour lines at this specified levels
             If array-like, draw contour lines at the specified levels.
             The values must be in increasing order.
            X: The coordinates of the values in *Z*.
             *X* must be 2-D with the same shape as *Z* (e.g. created via
             ``numpy.meshgrid``), or it must both be 1-D such that ``len(X) == M``
             is the number of columns in *Z*.
             If none, they are assumed to be integer indices, i.e. ``X = range(M)``.
            Y: The coordinates of the values in *Z*.
             *Y* must be 2-D with the same shape as *Z* (e.g. created via
             ``numpy.meshgrid``), or it must both be 1-D such that ``len(Y) == N``
             is the number of rows in *Z*.
             If none, they are assumed to be integer indices, i.e. ``Y = range(N)``.

        Returns:
            list of :py:class:`.ContourItem` objects
        """
        return create_contour_items(Z, levels, X, Y)

    def histogram2D(
        self,
        X: numpy.ndarray,
        Y: numpy.ndarray,
        NX: int | None = None,
        NY: int | None = None,
        logscale: bool | None = None,
        title: str | None = None,
        transparent: bool | None = None,
        Z: numpy.ndarray | None = None,
        computation: int = -1,
        interpolation: int = 0,
        lock_position: bool = True,
    ) -> Histogram2DItem:
        """Make a 2D Histogram `plot item`

        Args:
            X: X data
            Y: Y data
            NX: number of bins along x-axis. Default is None
            NY: number of bins along y-axis. Default is None
            logscale: Z-axis scale. Default is None
            title: item title. Default is None
            transparent: enable transparency. Default is None
            Z: Z data. Default is None
            computation: computation mode. Default is -1
            interpolation: interpolation mode. Default is 0
            lock_position: lock position. Default is True

        Returns:
            :py:class:`.Histogram2DItem` object
        """
        basename = _("2D Histogram")
        param = Histogram2DParam(title=basename, icon="histogram2d.png")
        if NX is not None:
            param.nx_bins = NX
        if NY is not None:
            param.ny_bins = NY
        if logscale is not None:
            param.logscale = int(logscale)
        if title is not None:
            param.label = title
        else:
            global HISTOGRAM2D_COUNT
            HISTOGRAM2D_COUNT += 1
            param.label = make_title(basename, HISTOGRAM2D_COUNT)
        if transparent is not None:
            param.transparent = transparent
        param.computation = computation
        param.interpolation = interpolation
        param.lock_position = lock_position
        return Histogram2DItem(X, Y, param, Z=Z)



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import TYPE_CHECKING

from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC
from qtpy import QtGui as QG

from plotpy.items.shape.polygon import PolygonShape

if TYPE_CHECKING:
    from plotpy.styles.shape import ShapeParam


class PointShape(PolygonShape):
    """Point shape

    Args:
        x: X coordinate
        y: Y coordinate
        shapeparam: Shape parameters
    """

    CLOSED = False
    _icon_name = "point_shape.png"

    def __init__(
        self, x: float = 0.0, y: float = 0.0, shapeparam: ShapeParam = None
    ) -> None:
        super().__init__(shapeparam=shapeparam)
        self.set_pos(x, y)

    def set_pos(self, x: float, y: float) -> None:
        """Set the point coordinates to (x, y)

        Args:
            x: X coordinate
            y: Y coordinate
        """
        self.set_points([(x, y)])

    def get_pos(self) -> tuple[float, float]:
        """Return the point coordinates

        Returns:
            Point coordinates
        """
        return tuple(self.points[0])

    def move_point_to(
        self, handle: int, pos: tuple[float, float], ctrl: bool = False
    ) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        nx, ny = pos
        self.points[0] = (nx, ny)

    def __reduce__(self) -> tuple:
        """Reduce object to picklable state"""
        state = (self.shapeparam, self.points, self.z())
        return (self.__class__, (), state)

    def __setstate__(self, state: tuple) -> None:
        """Set object state from pickled state"""
        self.shapeparam, self.points, z = state
        self.setZ(z)
        self.shapeparam.update_item(self)

    def boundingRect(self) -> QC.QRectF:
        """Return the bounding rectangle of the shape

        Returns:
            Bounding rectangle of the shape
        """
        poly = QG.QPolygonF()
        if self.ADDITIONNAL_POINTS:
            shape_points = self.points[: -self.ADDITIONNAL_POINTS]
        else:
            shape_points = self.points
        for i in range(shape_points.shape[0]):
            poly.append(QC.QPointF(shape_points[i, 0], shape_points[i, 1]))
        return poly.boundingRect()


assert_interfaces_valid(PointShape)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Curve plotting with time axis test"""

# guitest: show

from __future__ import annotations

import csv
from datetime import datetime, timedelta

import numpy as np
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.plot import PlotWidget
from plotpy.tests import get_path, vistools
from plotpy.tools import VCursorTool


def __create_time_data() -> tuple[np.ndarray, np.ndarray]:
    """Create time data"""
    # Create a temperature monitoring signal with second-resolution timestamps
    base_time = datetime(2025, 10, 7, 10, 0, 0)
    timestamps = [base_time + timedelta(seconds=i * 10) for i in range(100)]

    # Convert timestamps to numpy array of floats (seconds since epoch)
    x = np.array([(ts - datetime(1970, 1, 1)).total_seconds() for ts in timestamps])

    # Simulate temperature data with daily variation and noise
    hours_elapsed = np.arange(100) * 10 / 3600  # Convert to hours
    y = 20 + 5 * np.sin(2 * np.pi * hours_elapsed / 24) + np.random.randn(100) * 0.5

    return x, y


def test_plot_time_axis1():
    """Test plot time axis"""
    with qt_app_context(exec_loop=True):
        x, y = __create_time_data()
        items = [make.curve(x, y, color="b")]
        win = vistools.show_items(
            items, plot_type="curve", wintitle=test_plot_time_axis1.__doc__
        )
        win.manager.add_tool(VCursorTool)
        plot = win.manager.get_plot()
        # Configure x-axis for datetime display
        plot.set_axis_datetime("bottom", format="%H:%M:%S")
        # Set axis limits using datetime objects (zoom to first half of data)
        base_time = datetime(2025, 10, 7, 10, 0, 0)
        plot.set_axis_limits_from_datetime(
            "bottom",
            base_time,
            base_time + timedelta(minutes=10),
        )


def test_plot_time_axis2():
    """Test plot time axis with data from file"""
    datafile = get_path("datetime.txt")

    # Read data from file without Pandas dependency
    timestamps = []
    temperatures = []

    with open(datafile, "r", encoding="utf-8") as f:
        reader = csv.reader(f)
        next(reader)  # Skip header row

        for row in reader:
            if len(row) >= 3:  # Ensure we have at least timestamp and temperature
                try:
                    # Parse timestamp (column 1) and temperature (column 2)
                    timestamp_str = row[1]
                    temperature = float(row[2])

                    # Parse datetime and convert to seconds since epoch
                    dt = datetime.strptime(timestamp_str, "%Y-%m-%d %H:%M:%S")
                    timestamp_epoch = (dt - datetime(1970, 1, 1)).total_seconds()

                    timestamps.append(timestamp_epoch)
                    temperatures.append(temperature)
                except (ValueError, IndexError):
                    # Skip malformed rows
                    continue

    # Convert to numpy arrays
    x = np.array(timestamps)
    y = np.array(temperatures)

    with qt_app_context(exec_loop=True):
        items = [make.curve(x, y, color="r")]
        win = vistools.show_items(
            items, plot_type="curve", wintitle=test_plot_time_axis2.__doc__
        )
        plot = win.manager.get_plot()
        # Configure x-axis for datetime display
        plot.set_axis_datetime("bottom", format="%Y-%m-%d %H:%M")

        # Set axis limits using datetime objects
        if len(timestamps) > 0:
            dt1 = datetime(2025, 6, 19, 10, 0)
            dt2 = datetime(2025, 6, 19, 12, 0)
            plot.set_axis_limits_from_datetime("bottom", dt1, dt2)


def test_datetime_scale_type():
    """Test the new 'datetime' scale type API"""
    with qt_app_context():
        from plotpy.plot import PlotWidget

        plot_widget = PlotWidget()
        plot = plot_widget.plot

        # Test setting datetime scale type
        plot.set_axis_scale("bottom", "datetime")
        assert plot.get_axis_scale("bottom") == "datetime"

        # Test scale switching
        plot.set_axis_scale("bottom", "lin")
        assert plot.get_axis_scale("bottom") == "lin"

        plot.set_axis_scale("bottom", "log")
        assert plot.get_axis_scale("bottom") == "log"

        plot.set_axis_scale("bottom", "datetime")
        assert plot.get_axis_scale("bottom") == "datetime"

        # Verify the scale engine type
        from qwt import QwtDateTimeScaleEngine

        axis_id = plot.get_axis_id("bottom")
        engine = plot.axisScaleEngine(axis_id)
        assert isinstance(engine, QwtDateTimeScaleEngine)

        print("✓ Datetime scale type tests passed")


def test_datetime_coordinate_formatting():
    """Test that coordinates are formatted as datetime when axis is in datetime mode"""
    with qt_app_context():
        plot_widget = PlotWidget()
        plot = plot_widget.plot

        # Create datetime data
        start_time = datetime(2024, 1, 15, 10, 0, 0)
        times = [start_time + timedelta(hours=x) for x in range(24)]
        timestamps = [t.timestamp() for t in times]
        values = np.random.random(len(times))

        # Add curve and set datetime scale
        curve = make.curve(timestamps, values, title="Test curve")
        plot.add_item(curve)
        plot.set_axis_limits("bottom", timestamps[0], timestamps[-1])
        plot.set_axis_scale("bottom", "datetime")

        # Test coordinate formatting
        test_time = timestamps[12]  # Noon
        test_value = values[12]

        # Test plot-level coordinate display
        grid_coords = plot.get_coordinates_str(test_time, test_value)
        assert "2024-01-15" in grid_coords, (
            f"Grid coordinates should contain datetime: {grid_coords}"
        )

        # Test curve-level coordinate display
        curve_coords = curve.get_coordinates_label(test_time, test_value)
        assert "2024-01-15" in curve_coords, (
            f"Curve coordinates should contain datetime: {curve_coords}"
        )

        # Test with custom format
        plot.set_axis_datetime("bottom", format="%H:%M")
        custom_coords = plot.get_coordinates_str(test_time, test_value)
        expected_time = datetime.fromtimestamp(test_time).strftime("%H:%M")
        assert expected_time in custom_coords, (
            f"Custom format coordinates should contain {expected_time}: {custom_coords}"
        )

        # Test switch back to linear
        plot.set_axis_scale("bottom", "lin")
        linear_coords = plot.get_coordinates_str(test_time, test_value)
        assert (
            str(int(test_time)) in linear_coords or f"{test_time:g}" in linear_coords
        ), f"Linear coordinates should contain numeric timestamp: {linear_coords}"

        print("✓ Datetime coordinate formatting tests passed")


if __name__ == "__main__":
    test_plot_time_axis1()
    test_plot_time_axis2()
    test_datetime_scale_type()
    test_datetime_coordinate_formatting()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Test for to_bins function used in XYImageItem"""

import numpy as np

from plotpy.items.image.filter import to_bins


def test_to_bins_single_element():
    """Test to_bins with a single-element array.

    This is a regression test for the bug where loading an image with only
    1 row (e.g., SIF files with shape (1, N)) caused an IndexError because
    to_bins assumed at least 2 elements to compute bin edges.
    """
    x = np.array([5.0])
    result = to_bins(x)

    # Should return 2 bin edges, centered around the single point
    assert len(result) == 2
    assert result[0] == 4.5  # x[0] - 0.5
    assert result[1] == 5.5  # x[0] + 0.5


def test_to_bins_two_elements():
    """Test to_bins with a two-element array."""
    x = np.array([1.0, 3.0])
    result = to_bins(x)

    # Should return 3 bin edges
    assert len(result) == 3
    np.testing.assert_allclose(result, [0.0, 2.0, 4.0])


def test_to_bins_multiple_elements():
    """Test to_bins with multiple elements (uniform spacing)."""
    x = np.array([1.0, 2.0, 3.0, 4.0])
    result = to_bins(x)

    # Should return 5 bin edges
    assert len(result) == 5
    np.testing.assert_allclose(result, [0.5, 1.5, 2.5, 3.5, 4.5])


def test_to_bins_non_uniform_spacing():
    """Test to_bins with non-uniform spacing."""
    x = np.array([0.0, 1.0, 4.0])
    result = to_bins(x)

    # Should return 4 bin edges
    # First edge: 0.0 - (1.0 - 0.0) / 2 = -0.5
    # Middle: (0.0 + 1.0) / 2 = 0.5, (1.0 + 4.0) / 2 = 2.5
    # Last edge: 4.0 + (4.0 - 1.0) / 2 = 5.5
    assert len(result) == 4
    np.testing.assert_allclose(result, [-0.5, 0.5, 2.5, 5.5])


if __name__ == "__main__":
    test_to_bins_single_element()
    test_to_bins_two_elements()
    test_to_bins_multiple_elements()
    test_to_bins_non_uniform_spacing()
    print("All tests passed!")



from __future__ import annotations

import pytest
from guidata.env import execenv
from guidata.qthelpers import exec_dialog, qt_app_context

from plotpy.builder import make
from plotpy.tests.data import gen_image4
from plotpy.tests.unit.utils import (
    rel_pos_to_canvas_pos,
)
from plotpy.tools.image import ImageMaskTool
from plotpy.tools.shape import CircleTool, RectangleTool, RectangularShapeTool


@pytest.mark.parametrize("shape_tool_cls", [RectangleTool, CircleTool])
@pytest.mark.parametrize("inside", [True, False])
def test_image_mask_tool(shape_tool_cls: type[RectangularShapeTool], inside: bool):
    """Test the ImageMaskTool with a shape tool.

    Args:
        shape_tool_cls: The shape tool class to use.
        inside: If the mask should be inside or outside the shape.
    """
    with qt_app_context(exec_loop=False):
        win = make.dialog(type="image", toolbar=True)
        item = make.maskedimage(gen_image4(100, 100))
        plot = win.manager.get_plot()
        plot.add_item(item)  # type: ignore
        plot.select_item(item)  # type: ignore
        win.show()
        mask_tool = win.manager.add_tool(
            ImageMaskTool,
        )
        shape_tool = win.manager.add_tool(
            shape_tool_cls,
            toolbar_id=None,
            handle_final_shape_cb=lambda shape: mask_tool.handle_shape(
                shape, inside=inside
            ),
        )
        shape_tool.activate()
        xy0, xy1 = (0.1, 0.1), (0.9, 0.9)
        canvas = plot.canvas()
        assert canvas is not None
        pos0, _ = rel_pos_to_canvas_pos(canvas, xy0)
        pos1, _ = rel_pos_to_canvas_pos(canvas, xy1)

        with execenv.context(accept_dialogs=True):
            shape_tool.add_shape_to_plot(plot, pos0, pos1)
            mask_tool.apply_mask()

            # Clear mask
            item.unmask_all()
            plot.replot()

            shape_tool.add_shape_to_plot(plot, pos0, pos1)
            mask_tool.remove_shapes()

        exec_dialog(win)



# -*- coding: utf-8 -*-

from __future__ import annotations

import weakref
from typing import TYPE_CHECKING, Any

from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

from plotpy.config import CONF, _
from plotpy.constants import AXIS_IDS, LUT_MAX, X_BOTTOM, Y_LEFT, PlotType
from plotpy.interfaces import ICSImageItemType
from plotpy.panels.csection.csitem import (
    LineCrossSectionItem,
    ObliqueCrossSectionItem,
    XCrossSectionItem,
    YCrossSectionItem,
)
from plotpy.plot.base import BasePlot
from plotpy.styles.curve import CurveParam

LUT_AXIS_TITLE = _("LUT scale") + (" (0-%d)" % LUT_MAX)


if TYPE_CHECKING:
    from qtpy.QtWidgets import QWidget

    from plotpy.items import BaseImageItem
    from plotpy.panels.csection.csitem import CrossSectionItem


class BaseCrossSectionPlot(BasePlot):
    """Cross section plot base class

    Args:
        parent: Parent widget. Defaults to None.
    """

    CURVE_LABEL = _("Cross section")
    LABEL_TEXT = ""  # to be overridden in subclasses
    _HEIGHT = None
    _WIDTH = None
    CS_AXIS = None
    Z_AXIS = None
    Z_MAX_MAJOR = 5
    SHADE = 0.2

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(
            parent=parent,
            options=dict(title="", section="cross_section", type="curve"),
        )
        self.perimage_mode = True
        self.autoscale_mode = True
        self.autorefresh_mode = True
        self.apply_lut = False
        self.single_source = False
        self.lockscales = True

        self.last_obj = None
        self.known_items = {}
        self._shapes = {}

        self.param = CurveParam(_("Curve"), icon="curve.png")
        self.set_curve_style("cross_section", "curve")

        if self._HEIGHT is not None:
            self.setSizePolicy(QW.QSizePolicy.Expanding, QW.QSizePolicy.Minimum)
        elif self._WIDTH is not None:
            self.setSizePolicy(QW.QSizePolicy.Minimum, QW.QSizePolicy.Expanding)

        # The following import is here to avoid circular imports
        # pylint: disable=import-outside-toplevel
        from plotpy.builder import make

        self.label = make.label(self.LABEL_TEXT, "C", (0, 0), "C")
        self.label.set_readonly(True)
        self.add_item(self.label)

        self.setAxisMaxMajor(self.Z_AXIS, self.Z_MAX_MAJOR)
        self.setAxisMaxMinor(self.Z_AXIS, 0)

    def set_curve_style(self, section: str, option: str) -> None:
        """Set curve style

        Args:
            section: Configuration section name
            option: Configuration option name
        """
        self.param.read_config(CONF, section, option)
        self.param.label = self.CURVE_LABEL

    def connect_plot(self, plot: BasePlot) -> None:
        """Connect plot to cross section plot

        Args:
            plot: Plot to connect (containing the image items to be used for)
        """
        if plot.options.type == PlotType.CURVE:
            # Connecting only to image plot widgets (allow mixing image and
            # curve widgets for the same plot manager -- e.g. in pyplot)
            return
        plot.SIG_ITEMS_CHANGED.connect(self.items_changed)
        plot.SIG_LUT_CHANGED.connect(self.lut_changed)
        plot.SIG_MASK_CHANGED.connect(lambda item: self.update_plot())
        plot.SIG_ACTIVE_ITEM_CHANGED.connect(self.active_item_changed)
        plot.SIG_ANNOTATION_CHANGED.connect(self.shape_changed)
        plot.SIG_PLOT_LABELS_CHANGED.connect(self.plot_labels_changed)
        plot.SIG_AXIS_DIRECTION_CHANGED.connect(self.axis_dir_changed)
        plot.SIG_PLOT_AXIS_CHANGED.connect(self.plot_axis_changed)
        self.plot_labels_changed(plot)
        for axis_id in AXIS_IDS:
            self.axis_dir_changed(plot, axis_id)
        self.items_changed(plot)

    def register_shape(self, plot: BasePlot, shape: Any, refresh: bool = True) -> None:
        """Register shape associated to cross section plot

        Args:
            plot: Plot containing the shape and the associated image items
            shape: Shape to register, and to be used for cross section
            refresh: Whether to refresh the plot
        """
        known_shapes = self._shapes.get(plot, [])
        if shape in known_shapes:
            return
        self._shapes[plot] = known_shapes + [shape]
        self.update_plot(shape, refresh=refresh and self.autorefresh_mode)

    def unregister_shape(self, shape: Any) -> None:
        """Unregister shape associated to cross section plot

        Args:
            shape: Shape to unregister
        """
        for plot in self._shapes:
            shapes = self._shapes[plot]
            if shape in shapes:
                shapes.pop(shapes.index(shape))
                if len(shapes) == 0 or shape is self.get_last_obj():
                    for curve in self.get_cross_section_curves():
                        curve.clear_data()
                    self.replot()
                break

    def create_cross_section_item(self) -> CrossSectionItem:
        """Create cross section item"""
        raise NotImplementedError

    def add_cross_section_item(self, source: BaseImageItem) -> None:
        """Add cross section item

        Args:
            source: Source image item to add cross section item for
        """
        curve = self.create_cross_section_item()
        curve.set_source_image(source)
        curve.set_readonly(True)
        self.add_item(curve, z=0)
        self.known_items[source] = curve

    def get_cross_section_curves(self) -> list[CrossSectionItem]:
        """Return cross section curves"""
        return list(self.known_items.values())

    def items_changed(self, plot: BasePlot) -> None:
        """Items have changed in the plot

        Args:
            plot: Plot containing the image items to be used for cross section
        """
        # Del obsolete cross section items
        new_sources = plot.get_items(item_type=ICSImageItemType)
        for source in self.known_items.copy():
            if source not in new_sources:
                curve = self.known_items.pop(source)
                curve.clear_data()  # useful to emit SIG_CS_CURVE_CHANGED
                # (eventually notify other panels that the
                #  cross section curve is now empty)
                self.del_item(curve)

        # Update plot only to show/hide cross section curves according to
        # the associated image item visibility state (hence `refresh=False`)
        self.update_plot(refresh=False)

        self.plot_axis_changed(plot)

        if not new_sources:
            self.replot()
            return

        self.param.shade = self.SHADE / len(new_sources)
        for source in new_sources:
            if source not in self.known_items and source.isVisible():
                if not self.single_source or not self.known_items:
                    self.add_cross_section_item(source=source)

    def active_item_changed(self, plot: BasePlot) -> None:
        """Active item has just changed

        Args:
            plot: Plot containing the image items to be used for cross section
        """
        self.shape_changed(plot.get_active_item())

    def plot_labels_changed(self, plot: BasePlot) -> None:
        """Plot labels have changed

        Args:
            plot: Plot containing the image items to be used for cross section
        """
        raise NotImplementedError

    def axis_dir_changed(self, plot: BasePlot, axis_id: str) -> None:
        """An axis direction has changed

        Args:
            plot: Plot containing the image items to be used for cross section
            axis_id: Axis ID
        """
        pass

    def plot_axis_changed(self, plot: BasePlot) -> None:
        """Plot was just zoomed/panned

        Args:
            plot: Plot containing the image items to be used for cross section
        """
        if self.autoscale_mode:
            self.do_autoscale(replot=True)
        elif self.lockscales:
            self.do_autoscale(replot=False, axis_id=self.Z_AXIS)
            vmin, vmax = plot.get_axis_limits(self.CS_AXIS)
            self.set_axis_limits(self.CS_AXIS, vmin, vmax)
            self.replot()

    def is_shape_known(self, shape: Any) -> bool:
        """Return whether shape is known

        Args:
            shape: Shape to check

        Returns:
            bool: Whether shape is known
        """
        for shapes in list(self._shapes.values()):
            if shape in shapes:
                return True
        else:
            return False

    def shape_changed(self, shape: Any) -> None:
        """Shape has changed

        Args:
            shape: Shape that has changed
        """
        if self.autorefresh_mode:
            if self.is_shape_known(shape):
                self.update_plot(shape)

    def get_last_obj(self) -> BaseImageItem | None:
        """Return last active object"""
        if self.last_obj is not None:
            return self.last_obj()

    def update_plot(self, obj: Any = None, refresh: bool = True) -> None:
        """
        Update cross section curves

        Args:
            obj: Object to update cross section curves for
             (may be a marker or a rectangular shape, or None to update)
            refresh: Whether to refresh the plot
        """
        if obj is None:
            obj = self.get_last_obj()
            if obj is None:
                return
        else:
            self.last_obj = weakref.ref(obj)
        plot = obj.plot()
        if plot is None:
            self.unregister_shape(obj)
            return
        if self.label.isVisible():
            self.label.hide()
        items = list(self.known_items.items())
        for item, curve in iter(items):
            if not item.isVisible():
                curve.hide()
            else:
                curve.show()
                curve.perimage_mode = self.perimage_mode
                curve.autoscale_mode = self.autoscale_mode
                curve.apply_lut = self.apply_lut
                if refresh:
                    curve.update_item(obj)

        # Delegate all scaling logic to plot_axis_changed for consistency
        self.plot_axis_changed(plot)

    def toggle_perimage_mode(self, state: bool) -> None:
        """Toggle the per item mode

        Args:
            state: State to toggle to
        """
        self.perimage_mode = state
        self.update_plot()

    def toggle_autoscale(self, state: bool) -> None:
        """Toggle the autoscale mode

        Args:
            state: State to toggle to
        """
        self.autoscale_mode = state
        self.update_plot()

    def toggle_autorefresh(self, state: bool) -> None:
        """Toggle the autorefresh mode

        Args:
            state: State to toggle to
        """
        self.autorefresh_mode = state
        if state:
            self.update_plot()

    def toggle_apply_lut(self, state: bool) -> None:
        """Toggle the apply LUT mode

        Args:
            state: State to toggle to
        """
        self.apply_lut = state
        self.update_plot()
        if self.apply_lut:
            self.set_axis_title(self.Z_AXIS, LUT_AXIS_TITLE)
            self.set_axis_color(self.Z_AXIS, "red")
        else:
            obj = self.get_last_obj()
            if obj is not None and obj.plot() is not None:
                self.plot_labels_changed(obj.plot())

    def toggle_lockscales(self, state: bool) -> None:
        """Toggle the lock scales mode

        Args:
            state: State to toggle to
        """
        self.lockscales = state
        obj = self.get_last_obj()
        if obj is not None and obj.plot() is not None:
            self.plot_axis_changed(obj.plot())

    def lut_changed(self, plot: BasePlot) -> None:
        """LUT has changed

        Args:
            plot: Plot containing the image items to be used for cross section
        """
        if self.apply_lut:
            self.update_plot()


class HorizontalCrossSectionPlot(BaseCrossSectionPlot):
    LABEL_TEXT = _("Enable a marker")
    CS_AXIS = X_BOTTOM
    Z_AXIS = Y_LEFT

    def plot_labels_changed(self, plot: BasePlot) -> None:
        """Plot labels have changed

        Args:
            plot: Plot containing the image items to be used for cross section
        """
        self.set_axis_title("left", plot.get_axis_title("right"))
        self.set_axis_title("bottom", plot.get_axis_title("bottom"))
        self.set_axis_color("left", plot.get_axis_color("right"))
        self.set_axis_color("bottom", plot.get_axis_color("bottom"))

    def axis_dir_changed(self, plot: BasePlot, axis_id: str) -> None:
        """An axis direction has changed

        Args:
            plot: Plot containing the image items to be used for cross section
            axis_id: Axis ID
        """
        if axis_id == plot.X_BOTTOM:
            self.set_axis_direction("bottom", plot.get_axis_direction("bottom"))
            self.replot()


class VerticalCrossSectionPlot(BaseCrossSectionPlot):
    LABEL_TEXT = _("Enable a marker")
    CS_AXIS = Y_LEFT
    Z_AXIS = X_BOTTOM
    Z_MAX_MAJOR = 3

    def plot_labels_changed(self, plot: BasePlot) -> None:
        """Plot labels have changed

        Args:
            plot: Plot containing the image items to be used for cross section
        """
        self.set_axis_title("bottom", plot.get_axis_title("right"))
        self.set_axis_title("left", plot.get_axis_title("left"))
        self.set_axis_color("bottom", plot.get_axis_color("right"))
        self.set_axis_color("left", plot.get_axis_color("left"))

    def axis_dir_changed(self, plot: BasePlot, axis_id: str) -> None:
        """An axis direction has changed

        Args:
            plot: Plot containing the image items to be used for cross section
            axis_id: Axis ID
        """
        if axis_id == plot.Y_LEFT:
            self.set_axis_direction("left", plot.get_axis_direction("left"))
            self.replot()


class XYCrossSectionMixin:
    def connect_plot(self, plot: BasePlot) -> None:
        """Connect plot to cross section plot

        Args:
            plot: Plot to connect (containing the image items to be used for)
        """
        BaseCrossSectionPlot.connect_plot(self, plot)
        plot.SIG_MARKER_CHANGED.connect(
            lambda marker: BaseCrossSectionPlot.update_plot(self, marker)
        )


class XCrossSectionPlot(HorizontalCrossSectionPlot, XYCrossSectionMixin):
    """X-axis cross section plot"""

    _HEIGHT = 130

    def sizeHint(self) -> QC.QSize:
        """Reimplemented from QWidget.sizeHint"""
        return QC.QSize(self.width(), self._HEIGHT)

    def create_cross_section_item(self) -> XCrossSectionItem:
        """Create cross section item"""
        return XCrossSectionItem(self.param)


class YCrossSectionPlot(VerticalCrossSectionPlot, XYCrossSectionMixin):
    """Y-axis cross section plot"""

    _WIDTH = 140

    def sizeHint(self) -> QC.QSize:
        """Reimplemented from QWidget.sizeHint"""
        return QC.QSize(self._WIDTH, self.height())

    def create_cross_section_item(self) -> YCrossSectionItem:
        """Create cross section item"""
        return YCrossSectionItem(self.param)


# Oblique cross section plot
class ObliqueCrossSectionPlot(HorizontalCrossSectionPlot):
    """Oblique averaged cross section plot"""

    PLOT_TITLE = _("Oblique averaged cross section")
    CURVE_LABEL = _("Oblique averaged cross section")
    LABEL_TEXT = _("Activate the oblique cross section tool")
    SHADE = 0.0

    def __init__(self, parent=None):
        super().__init__(parent)
        self.set_title(self.PLOT_TITLE)
        self.single_source = True

    def create_cross_section_item(self) -> ObliqueCrossSectionItem:
        """Create cross section item"""
        return ObliqueCrossSectionItem(self.param)


# Line cross section plot
class LineCrossSectionPlot(HorizontalCrossSectionPlot):
    """Line cross section plot"""

    PLOT_TITLE = _("Line cross section")
    CURVE_LABEL = _("Line cross section")
    LABEL_TEXT = _("Activate the line cross section tool")
    SHADE = 0.0

    def __init__(self, parent=None):
        super().__init__(parent)
        self.set_title(self.PLOT_TITLE)
        self.single_source = True

    def create_cross_section_item(self) -> LineCrossSectionItem:
        """Create cross section item"""
        return LineCrossSectionItem(self.param)



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pytest
import qtpy.QtCore as QC
from guidata.qthelpers import qt_app_context

from plotpy.interfaces.items import ICurveItemType, IImageItemType
from plotpy.tests.unit.utils import create_window, drag_mouse
from plotpy.tools import DisplayCoordsTool

if TYPE_CHECKING:
    from plotpy.interfaces.items import IItemType


@pytest.mark.parametrize("active_item", [ICurveItemType, IImageItemType, None])
def test_display_coords(active_item: type[IItemType] | None):
    """Test display coordinates tool on a curve and on an image."""
    with qt_app_context(exec_loop=False):
        win, _tool = create_window(DisplayCoordsTool, active_item_type=active_item)
        plot = win.manager.get_plot()

        # The is no way to test a condition while the mouse is moving so it is
        # not possible to test the display of the coordinates while the mouse is moving.
        assert plot.curve_pointer is False and plot.canvas_pointer is False
        drag_mouse(win, np.array([0.5]), np.array([0.5]), click=False)
        assert plot.curve_pointer is False and plot.canvas_pointer is False
        drag_mouse(win, np.array([0.5]), np.array([0.5]), click=True)
        assert plot.curve_pointer is False and plot.canvas_pointer is False
        drag_mouse(
            win,
            np.array([0.5]),
            np.array([0.5]),
            click=False,
            mod=QC.Qt.KeyboardModifier.AltModifier,
        )
        assert plot.curve_pointer is False and plot.canvas_pointer is False
        drag_mouse(
            win,
            np.array([0.5]),
            np.array([0.5]),
            click=False,
            mod=QC.Qt.KeyboardModifier.ControlModifier,
        )
        assert plot.curve_pointer is False and plot.canvas_pointer is False


if __name__ == "__main__":
    for item in (ICurveItemType, IImageItemType):
        test_display_coords(item)



from qtpy.QtCore import Signal

from ._generic_range_slider import _GenericRangeSlider
from ._generic_slider import _GenericSlider


class _IntMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._singleStep = 1

    def _type_cast(self, value) -> int:
        return int(round(value))


class _FloatMixin:
    _fvalueChanged = Signal(float)
    _fsliderMoved = Signal(float)
    _frangeChanged = Signal(float, float)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._singleStep = 0.01
        self._pageStep = 0.1

    def _type_cast(self, value) -> float:
        return float(value)


class QDoubleSlider(_FloatMixin, _GenericSlider):
    pass


class QIntSlider(_IntMixin, _GenericSlider):
    # mostly just an example... use QSlider instead.
    valueChanged = Signal(int)


class QRangeSlider(_IntMixin, _GenericRangeSlider):
    pass


class QDoubleRangeSlider(_FloatMixin, QRangeSlider):
    pass


# QRangeSlider.__doc__ += "\n" + textwrap.indent(QSlider.__doc__, "    ")



# -*- coding: utf-8 -*-

from qtpy import QtCore as QC

from plotpy.config import _
from plotpy.constants import SHAPE_Z_OFFSET
from plotpy.events import QtDragHandler, setup_standard_tool_filter
from plotpy.items import Marker, XRangeSelection
from plotpy.tools.base import DefaultToolbarID, InteractiveTool


class BaseCursorTool(InteractiveTool):
    """ """

    TITLE = None
    ICON = None

    def __init__(
        self,
        manager,
        toolbar_id=DefaultToolbarID,
        title=None,
        icon=None,
        tip=None,
        switch_to_default_tool=None,
    ):
        super().__init__(
            manager,
            toolbar_id,
            title=title,
            icon=icon,
            tip=tip,
            switch_to_default_tool=switch_to_default_tool,
        )
        self.shape = None

    def create_shape(self):
        """Create and return the cursor/range shape"""
        raise NotImplementedError

    def setup_filter(self, baseplot):
        """

        :param baseplot:
        :return:
        """
        filter = baseplot.filter
        # Initialisation du filtre
        start_state = filter.new_state()
        # Bouton gauche :
        self.handler = QtDragHandler(
            filter, QC.Qt.MouseButton.LeftButton, start_state=start_state
        )
        self.handler.SIG_MOVE.connect(self.move)
        self.handler.SIG_STOP_NOT_MOVING.connect(self.end_move)
        self.handler.SIG_STOP_MOVING.connect(self.end_move)
        return setup_standard_tool_filter(filter, start_state)

    def move(self, filter, event):
        """

        :param filter:
        :param event:
        """
        plot = filter.plot
        if not self.shape:
            self.shape = self.create_shape()
            self.shape.attach(plot)
            self.shape.setZ(plot.get_max_z() + 1)
            self.shape.move_local_point_to(0, event.pos())
            self.shape.setVisible(True)
        self.shape.move_local_point_to(1, event.pos())
        plot.replot()

    def end_move(self, filter, event):
        """

        :param filter:
        :param event:
        """
        if self.shape is not None:
            assert self.shape.plot() == filter.plot
            filter.plot.add_item_with_z_offset(self.shape, SHAPE_Z_OFFSET)
            self.shape = None
            self.SIG_TOOL_JOB_FINISHED.emit()


class HRangeTool(BaseCursorTool):
    TITLE = _("Horizontal selection")
    ICON = "xrange.png"

    def create_shape(self):
        """

        :return:
        """
        return XRangeSelection(0, 0)


class VCursorTool(BaseCursorTool):
    TITLE = _("Vertical cursor")
    ICON = "vcursor.png"

    def create_shape(self):
        """

        :return:
        """
        marker = Marker()
        marker.set_markerstyle("|")
        return marker


class HCursorTool(BaseCursorTool):
    TITLE = _("Horizontal cursor")
    ICON = "hcursor.png"

    def create_shape(self):
        """

        :return:
        """
        marker = Marker()
        marker.set_markerstyle("-")
        return marker


class XCursorTool(BaseCursorTool):
    TITLE = _("Cross cursor")
    ICON = "xcursor.png"

    def create_shape(self):
        """

        :return:
        """
        marker = Marker()
        marker.set_markerstyle("+")
        return marker



# script/reinstall_dev.py
"""
Reinstall multiple local libraries in editable mode for development.

Workflow:
  1) Try to uninstall all target libraries in one command (ignore errors if some are not installed).
  2) Reinstall each library in editable mode from a sibling folder: ../<library>.

This script uses the same Python interpreter that runs it (sys.executable),
so pip operations happen in the same environment (e.g., your active venv).
"""

from __future__ import annotations

import os
import shutil
import subprocess
import sys
import sysconfig
from typing import Iterable

# Absolute path to the Python executable that runs this script.
# Using sys.executable ensures pip targets the same environment (e.g., your venv).
PY = sys.executable


def run(cmd: list[str]) -> None:
    """
    Echo and execute a subprocess command.
    Raises:
        subprocess.CalledProcessError: if the command returns a non-zero exit code.
    """
    print("$", " ".join(cmd), flush=True)
    subprocess.check_call(cmd)


def uninstall_many(packages: Iterable[str]) -> None:
    """
    Attempt to uninstall multiple packages at once.
    If uninstall fails (e.g., a package is not installed), we keep going.

    Args:
        packages: An iterable of package names (pip distribution names).
    """
    pkgs = list(packages)
    if not pkgs:
        return
    try:
        run([PY, "-m", "pip", "uninstall", "-y", *pkgs])
    except subprocess.CalledProcessError as e:
        # Continue even if the uninstall step fails (for one or more packages)
        print(f"[WARN] Uninstall returned {e.returncode} — continuing...", flush=True)


def install_editable_many(packages: Iterable[str]) -> None:
    """
    Install each package in editable mode from ../<package_dir>.

    Assumes your project layout has sibling folders one level up, e.g.:
      ../guidata

    Args:
        packages: An iterable of package names (also used as directory names).
    """
    for pkg in packages:
        run([PY, "-m", "pip", "install", "-e", f"../{pkg}"])


def remove_residual_dirs(packages: Iterable[str]) -> None:
    """
    Force remove residual package directories from site-packages.
    This mimics: rm -rf .venv/Lib/site-packages/<pkg>
    to ensure a clean slate before reinstalling.
    """
    # Locates site-packages, e.g. .venv/Lib/site-packages
    site_packages = sysconfig.get_path("purelib")

    for pkg in packages:
        target = os.path.join(site_packages, pkg)
        if os.path.isdir(target):
            print(f"Removing residual directory: {target}", flush=True)
            try:
                shutil.rmtree(target)
            except OSError as e:
                print(f"[WARN] Failed to remove {target}: {e}", flush=True)


def reinstall_packages(packages: list[str]) -> None:
    """
    High-level orchestration for many packages:
      - Uninstall all of them (ignore failures)
      - Force remove residual directories from site-packages
      - Install each in editable mode
    """
    # 1) Uninstall (ignore if not installed)
    uninstall_many(packages)

    # 2) Force remove residual directories (essential for clean reinstall)
    remove_residual_dirs(packages)

    # 3) Editable installs
    install_editable_many(packages)


if __name__ == "__main__":
    # ⭐ Fixed, editable local libraries to manage (edit as needed)
    PACKAGES = ["guidata"]

    print("🏃 Reinstalling editable packages:", ", ".join(PACKAGES))
    reinstall_packages(PACKAGES)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

# pylint: disable=C0103

from __future__ import annotations

from sys import maxsize
from typing import TYPE_CHECKING

import numpy as np
from guidata.dataset import update_dataset
from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qwt import QwtPlotItem

from plotpy.config import _
from plotpy.interfaces import IBasePlotItem, ISerializableType, ITrackableItemType
from plotpy.styles import PolygonMapParam

if TYPE_CHECKING:
    import guidata.io
    from qtpy.QtCore import QPointF
    from qwt import QwtScaleMap

    from plotpy.interfaces import IItemType
    from plotpy.styles.base import ItemParameters


def simplify_poly(pts, off, scale, bounds):
    """Simplify a polygon by removing points outside the canvas"""
    ax, bx, ay, by = scale
    a = np.array([[ax, ay]])
    b = np.array([[bx, by]])
    _pts = a * pts + b
    poly = []
    NP = off.shape[0]
    for i in range(off.shape[0]):
        i0 = off[i, 1]
        if i + 1 < NP:
            i1 = off[i + 1, 1]
        else:
            i1 = pts.shape[0]
        poly.append((_pts[i0:i1], i))
    return poly


class PolygonMapItem(QwtPlotItem):
    """Construct a PolygonMapItem object

    Args:
        param: Polygon parameters
    """

    __implements__ = (IBasePlotItem, ISerializableType)

    _readonly = False
    _private = False
    _can_select = False
    _can_resize = False
    _can_move = False
    _can_rotate = False
    _icon_name = "polygonmap.png"

    def __init__(self, param: PolygonMapParam | None = None) -> None:
        super().__init__()
        if param is None:
            self.param = PolygonMapParam(_("PolygonMap"), icon="polygonmap.png")
        else:
            self.param = param
        self.selected = False
        self.immutable = True  # set to false to allow moving points around
        self._pts = None  # Array of points Mx2
        self._n = None  # Array of polygon offsets/ends Nx1
        #                 (polygon k points are _pts[_n[k-1]:_n[k]])
        self._c = None  # Color of polygon Nx2 [border,background] as RGBA uint32
        self.update_params()

    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (ITrackableItemType, ISerializableType)

    def can_select(self) -> bool:
        """
        Returns True if this item can be selected

        Returns:
            bool: True if item can be selected, False otherwise
        """
        return self._can_select

    def can_resize(self) -> bool:
        """
        Returns True if this item can be resized

        Returns:
            bool: True if item can be resized, False otherwise
        """
        return self._can_resize

    def can_rotate(self) -> bool:
        """
        Returns True if this item can be rotated

        Returns:
            bool: True if item can be rotated, False otherwise
        """
        return self._can_rotate

    def can_move(self) -> bool:
        """
        Returns True if this item can be moved

        Returns:
            bool: True if item can be moved, False otherwise
        """
        return self._can_move

    def get_icon_name(self) -> str:
        """Return the icon name

        Returns:
            Icon name
        """
        return self._icon_name

    def set_icon_name(self, icon_name: str) -> None:
        """Set the icon name

        Args:
            icon_name: Icon name
        """
        self._icon_name = icon_name

    def set_selectable(self, state: bool) -> None:
        """Set item selectable state

        Args:
            state: True if item is selectable, False otherwise
        """
        self._can_select = state

    def set_resizable(self, state: bool) -> None:
        """Set item resizable state
        (or any action triggered when moving an handle, e.g. rotation)

        Args:
            state: True if item is resizable, False otherwise
        """
        self._can_resize = state

    def set_movable(self, state: bool) -> None:
        """Set item movable state

        Args:
            state: True if item is movable, False otherwise
        """
        self._can_move = state

    def set_rotatable(self, state: bool) -> None:
        """Set item rotatable state

        Args:
            state: True if item is rotatable, False otherwise
        """
        self._can_rotate = state

    def __reduce__(self):
        state = (self.param, self._pts, self._n, self._c, self.z())
        res = (PolygonMapItem, (), state)
        return res

    def __setstate__(self, state):
        param, pts, n, c, z = state
        self.param = param
        self.set_data(pts, n, c)
        self.setZ(z)
        self.update_params()

    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        writer.write(self._pts, group_name="Pdata")
        writer.write(self._n, group_name="Ndata")
        writer.write(self._c, group_name="Cdata")
        writer.write(self.z(), group_name="z")
        self.param.update_param(self)
        writer.write(self.param, group_name="param")

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        pts = reader.read(group_name="Pdata", func=reader.read_array)
        n = reader.read(group_name="Ndata", func=reader.read_array)
        c = reader.read(group_name="Cdata", func=reader.read_array)
        self.set_data(pts, n, c)
        self.setZ(reader.read("z"))
        self.param = PolygonMapParam(_("PolygonMap"), icon="polygonmap.png")
        reader.read("param", instance=self.param)
        self.update_params()

    def set_readonly(self, state: bool) -> None:
        """Set object readonly state

        Args:
            state: True if object is readonly, False otherwise
        """
        self._readonly = state

    def is_readonly(self) -> bool:
        """Return object readonly state

        Returns:
            bool: True if object is readonly, False otherwise
        """
        return self._readonly

    def set_private(self, state: bool) -> None:
        """Set object as private

        Args:
            state: True if object is private, False otherwise
        """
        self._private = state

    def is_private(self) -> bool:
        """Return True if object is private

        Returns:
            bool: True if object is private, False otherwise
        """
        return self._private

    def invalidate_plot(self) -> None:
        """Invalidate the plot to force a redraw"""
        plot = self.plot()
        if plot is not None:
            plot.invalidate()

    def select(self) -> None:
        """
        Select the object and eventually change its appearance to highlight the
        fact that it's selected
        """

    def unselect(self) -> None:
        """
        Unselect the object and eventually restore its original appearance to
        highlight the fact that it's not selected anymore
        """

    def get_data(self):
        """Return curve data x, y (NumPy arrays)"""
        return self._pts, self._n, self._c

    def set_data(self, pts, n, c):
        """
        Set curve data:
            * x: NumPy array
            * y: NumPy array
        """
        self._pts = np.asarray(pts)
        self._n = np.asarray(n)
        self._c = np.asarray(c)
        xmin, ymin = self._pts.min(axis=0)
        xmax, ymax = self._pts.max(axis=0)
        self.bounds = QC.QRectF(xmin, ymin, xmax - xmin, ymax - ymin)

    def is_empty(self) -> bool:
        """Return True if the item is empty

        Returns:
            True if the item is empty, False otherwise
        """
        return self._pts is None or self._pts.size == 0

    def hit_test(self, pos: QPointF) -> tuple[float, float, bool, None]:
        """Return a tuple (distance, attach point, inside, other_object)

        Args:
            pos: Position

        Returns:
            tuple: Tuple with four elements: (distance, attach point, inside,
             other_object).

        Description of the returned values:

        * distance: distance in pixels (canvas coordinates) to the closest
           attach point
        * attach point: handle of the attach point
        * inside: True if the mouse button has been clicked inside the object
        * other_object: if not None, reference of the object which will be
           considered as hit instead of self
        """
        if self.is_empty():
            return maxsize, 0, False, None
        # TODO: Implement PolygonMapItem.hit_test
        return maxsize, 0, False, None

    def get_closest_coordinates(self, x: float, y: float) -> tuple[float, float]:
        """
        Get the closest coordinates to the given point

        Args:
            x: X coordinate
            y: Y coordinate

        Returns:
            tuple[float, float]: Closest coordinates
        """
        # TODO: Implement PolygonMapItem.get_closest_coordinates
        return x, y

    def get_coordinates_label(self, x: float, y: float) -> str:
        """
        Get the coordinates label for the given coordinates

        Args:
            x: X coordinate
            y: Y coordinate

        Returns:
            str: Coordinates label
        """
        title = self.title().text()
        return f"{title}:<br>x = {x:f}<br>y = {y:f}"

    def move_local_point_to(self, handle: int, pos: QPointF, ctrl: bool = None) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """

    def move_local_shape(self, old_pos: QPointF, new_pos: QPointF) -> None:
        """Translate the shape such that old_pos becomes new_pos in canvas coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """

    def move_with_selection(self, delta_x: float, delta_y: float) -> None:
        """Translate the item together with other selected items

        Args:
            delta_x: Translation in plot coordinates along x-axis
            delta_y: Translation in plot coordinates along y-axis
        """

    def update_params(self) -> None:
        """Update object properties from item parameters (dataset)"""
        self.param.update_item(self)
        if self.selected:
            self.select()

    def update_item_parameters(self) -> None:
        """Update item parameters (dataset) from object properties"""
        self.param.update_param(self)

    def get_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Appends datasets to the list of DataSets describing the parameters
        used to customize apearance of this item

        Args:
            itemparams: Item parameters
        """
        itemparams.add("CurveParam", self, self.param)

    def set_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Change the appearance of this item according
        to the parameter set provided

        Args:
            itemparams: Item parameters
        """
        update_dataset(self.param, itemparams.get("PolygonMapParam"), visible_only=True)
        self.update_params()

    def draw(
        self,
        painter: QG.QPainter,
        xMap: QwtScaleMap,
        yMap: QwtScaleMap,
        canvasRect: QC.QRectF,
    ) -> None:
        """Draw the item

        Args:
            painter: Painter
            xMap: X axis scale map
            yMap: Y axis scale map
            canvasRect: Canvas rectangle
        """
        p1x = xMap.p1()
        s1x = xMap.s1()
        ax = (xMap.p2() - p1x) / (xMap.s2() - s1x)
        p1y = yMap.p1()
        s1y = yMap.s1()
        ay = (yMap.p2() - p1y) / (yMap.s2() - s1y)
        bx, by = p1x - s1x * ax, p1y - s1y * ay
        _c = self._c
        _n = self._n
        fgcol = QG.QColor()
        bgcol = QG.QColor()
        polygons = simplify_poly(
            self._pts, _n, (ax, bx, ay, by), canvasRect.getCoords()
        )
        for poly, num in polygons:
            points = []
            for i in range(poly.shape[0]):
                points.append(QC.QPointF(poly[i, 0], poly[i, 1]))
            pg = QG.QPolygonF(points)
            fgcol.setRgba(int(_c[num, 0]))
            bgcol.setRgba(int(_c[num, 1]))
            painter.setPen(QG.QPen(fgcol))
            painter.setBrush(QG.QBrush(bgcol))
            painter.drawPolygon(pg)

    def boundingRect(self) -> QC.QRectF:
        """Return the bounding rectangle of the shape

        Returns:
            Bounding rectangle of the shape
        """
        return self.bounds


assert_interfaces_valid(PolygonMapItem)



# -*- coding: utf-8 -*-

# pylint: disable=unused-import
# flake8: noqa

from .items import (
    IBaseImageItem,
    IBasePlotItem,
    IColormapImageItemType,
    ICSImageItemType,
    ICurveItemType,
    IDecoratorItemType,
    IExportROIImageItemType,
    IHistDataSource,
    IImageItemType,
    IItemType,
    ISerializableType,
    IShapeItemType,
    ITrackableItemType,
    IVoiImageItemType,
)
from .panel import IPanel
from .plotmanager import IPlotManager



class IPlotManager:
    """A 'controller' that organizes relations between
    plots (BasePlot), panels, tools (GuiTool) and toolbar
    """

    def add_plot(self, plot, plot_id="default"):
        """

        :param plot:
        :param plot_id:
        """
        # Avoid circular import
        # pylint: disable=import-outside-toplevel
        from plotpy.plot.base import BasePlot

        assert id not in self.plots  # pylint: disable=no-member
        assert isinstance(plot, BasePlot)

    def add_panel(self, panel):
        """

        :param panel:
        """
        assert id not in self.panels  # pylint: disable=no-member

    def add_toolbar(self, toolbar, toolbar_id="default"):
        """

        :param toolbar:
        :param toolbar_id:
        """
        assert id not in self.toolbars  # pylint: disable=no-member

    def get_active_plot(self):
        """The active plot is the plot whose canvas has the focus
        otherwise it's the "default" plot
        """
        pass



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pytest
from guidata.qthelpers import exec_dialog, qt_app_context

from plotpy.tests.unit.utils import (
    create_window,
    drag_mouse,
)
from plotpy.tools import (
    HCursorTool,
    HRangeTool,
    VCursorTool,
    XCursorTool,
)

if TYPE_CHECKING:
    from plotpy.tools.cursor import BaseCursorTool


@pytest.mark.parametrize(
    "cursor_tool", [HCursorTool, VCursorTool, XCursorTool, HRangeTool]
)
def test_cursor_tool(cursor_tool: type[BaseCursorTool]):
    """Test the cursor tools. by simulating a mouse drag and checking if the tool
    shape is created.

    Args:
        cursor_tool: Cursor tool class to test.
    """
    with qt_app_context():
        win, _tool = create_window(cursor_tool)
        win.show()
        plot = win.manager.get_plot()

        active_tool = win.manager.get_active_tool()
        assert isinstance(active_tool, cursor_tool)
        tool_shape_type = type(active_tool.create_shape())
        assert tool_shape_type not in (type(item) for item in plot.get_items())

        drag_mouse(win, np.array([0.5, 0.6, 0.7]), np.array([0.5, 0.6, 0.7]))
        assert tool_shape_type in (type(item) for item in plot.get_items())

        exec_dialog(win)


if __name__ == "__main__":
    for tool in (HCursorTool, VCursorTool, XCursorTool, HRangeTool):
        test_cursor_tool(tool)
